from __future__ import annotations
import collections.abc
import contextlib
import threading
import tempfile
import weakref
import logging
import pickle
import shelve
import uuid
import os
import io

import claripy

from .errors import AngrVaultError
from .project import Project
from .sim_state import SimState
from .sim_type import SimType

l = logging.getLogger("angr.vault")


class VaultPickler(pickle.Pickler):
    def __init__(self, vault, file, *args, assigned_objects=(), **kwargs):
        """
        A persistence-aware pickler.
        It will check for persistence of any objects except for those with IDs in 'assigned_objects'.
        """

        super().__init__(file, *args, **kwargs)
        self.vault = vault
        self.assigned_objects = assigned_objects

    def persistent_id(self, obj):
        if any(obj is o for o in self.assigned_objects):
            return None

        pid = self.vault._get_persistent_id(obj)
        if pid is None:
            return None

        # l.debug("Persistent store: %s %s", obj, pid)
        return self.vault._store(obj, pid)


class VaultUnpickler(pickle.Unpickler):
    def __init__(self, vault, file, *args, **kwargs):
        super().__init__(file, *args, **kwargs)
        self.vault = vault

    def persistent_load(self, pid):
        return self.vault._load(pid)


class Vault(collections.abc.MutableMapping):
    """
    The vault is a serializer for angr.
    """

    #
    # These MUST be overridden.
    #

    def _read_context(self, i):
        """
        Should be a context that yields a pickle-read()able file object for the given id i.
        """
        raise NotImplementedError

    def _write_context(self, i):
        """
        Should be a context that yields a pickle-write()able file object for the given id i.
        """
        raise NotImplementedError

    def keys(self):
        """
        Should return the IDs stored by the vault.
        """
        raise NotImplementedError

    #
    # Persistence managers
    #

    def __init__(self):
        self._object_cache = weakref.WeakValueDictionary()
        self._uuid_cache = weakref.WeakKeyDictionary()
        self.stored = set()
        self.storing = set()
        self.hash_dedup = {
            claripy.ast.Base,
            claripy.ast.BV,
            claripy.ast.FP,
            claripy.ast.Bool,
            claripy.ast.Bits,
        }
        self.module_dedup = set()  # {'claripy', 'angr', 'archinfo', 'pyvex' } # cle causes recursion
        self.uuid_dedup = {SimState, Project}
        self.unsafe_key_baseclasses = {claripy.ast.Base, SimType}

    def _get_persistent_id(self, o):
        """
        Determines a persistent ID for an object.
        Does NOT do stores.
        """
        if type(o) in self.hash_dedup:
            oid = o.__class__.__name__ + "-" + str(hash(o))
            self._object_cache[oid] = o
            return oid

        if any(isinstance(o, c) for c in self.unsafe_key_baseclasses):
            return None

        try:
            return self._uuid_cache[o]
        except KeyError:
            pass
        except TypeError:
            return None

        # if type(o) in self.uuid_dedup:
        #    return self._get_id(o)
        if o.__class__.__module__.split(".")[0] in self.module_dedup or o.__class__ in self.uuid_dedup:
            oid = o.__class__.__name__.split(".")[-1] + "-" + str(uuid.uuid4())
            self._object_cache[oid] = o
            self._uuid_cache[o] = oid
            return oid

        return None

    #
    # Other stuff
    #

    def is_stored(self, i):
        """
        Checks if the provided id is already in the vault.
        """
        if i in self.stored:
            return True

        try:
            with self._read_context(i):
                return True
        except (AngrVaultError, EOFError):
            return False

    def load(self, oid):
        return self._load(oid)

    def _load(self, oid):
        """
        Retrieves one object from the pickler with the provided id.

        :param oid: an ID to use
        """
        # l.debug("LOAD: %s", oid)
        try:
            # l.debug("... trying cached")
            return self._object_cache[oid]
        except KeyError:
            # l.debug("... cached failed")
            with self._read_context(oid) as u:
                # add newly loaded object into the object cache
                o = VaultUnpickler(self, u).load()
                self._object_cache[oid] = o
                return o

    def store(self, o):
        actual_id = self._get_persistent_id(o) or "TMP-" + str(uuid.uuid4())

        return self._store(o, actual_id)

    def _store(self, o, oid):  # pylint:disable=redefined-builtin
        """
        Stores an object and returns its ID.

        :param o: the object
        :param oid: an ID to use
        """

        actual_id = oid

        # l.debug("STORE: %s %s", o, actual_id)

        # this handles recursive objects
        if actual_id in self.storing:
            return actual_id

        if self.is_stored(actual_id):
            # l.debug("... already stored")
            return actual_id

        with self._write_context(actual_id) as output:
            self.storing.add(actual_id)
            VaultPickler(self, output, assigned_objects=(o,)).dump(o)
            self.stored.add(actual_id)

        return actual_id

    def dumps(self, o):
        """
        Returns a serialized string representing the object, post-deduplication.

        :param o: the object
        """
        f = io.BytesIO()
        VaultPickler(self, f).dump(o)
        f.seek(0)
        return f.read()

    def loads(self, s):
        """
        Deserializes a string representation of the object.

        :param s: the string
        """
        f = io.BytesIO(s)
        return VaultUnpickler(self, f).load()

    def _clear_cache(self):
        self._object_cache.clear()
        self._uuid_cache.clear()
        self.stored.clear()
        self.storing.clear()

    @staticmethod
    def close():
        pass

    #
    # For MutableMapping.
    #

    def __setitem__(self, k, v):
        self._store(v, k)

    def __getitem__(self, k):
        return self.load(k)

    def __delitem__(self, k):
        raise AngrVaultError("We currently don't support deletion from the vault.")

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())


class VaultDict(Vault):
    """
    A Vault that uses a dictionary for storage.
    """

    def __init__(self, d=None):
        super().__init__()
        self._dict = {} if d is None else d

    @contextlib.contextmanager
    def _write_context(self, i):
        f = io.BytesIO()
        yield f
        f.seek(0)
        self._dict[i] = f.read()

    @contextlib.contextmanager
    def _read_context(self, i):
        try:
            f = io.BytesIO(self._dict[i])
            yield f
        except KeyError as e:
            raise AngrVaultError from e

    def is_stored(self, i):
        return i in self._dict

    def keys(self):
        return self._dict.keys()


class VaultDir(Vault):
    """
    A Vault that uses a directory for storage.
    """

    def __init__(self, d=None):
        super().__init__()
        self._dir = tempfile.mkdtemp() if d is None else d
        with contextlib.suppress(FileExistsError):
            os.makedirs(self._dir)

    @contextlib.contextmanager
    def _write_context(self, i):
        with open(os.path.join(self._dir, i), "wb") as o:
            yield o

    @contextlib.contextmanager
    def _read_context(self, i):
        try:
            with open(os.path.join(self._dir, i), "rb") as o:
                yield o
        except FileNotFoundError as e:
            raise AngrVaultError from e

    def keys(self):
        return os.listdir(self._dir)


class VaultShelf(VaultDict):
    """
    A Vault that uses a shelve.Shelf for storage.
    """

    def __init__(self, path=None):
        self._path = tempfile.mktemp() if path is None else path
        s = shelve.open(self._path, protocol=-1)  # noqa: SIM115
        super().__init__(s)

    def close(self):
        self._dict.close()


class VaultDirShelf(VaultDict):
    """
    A Vault that uses a directory for storage, where every object is stored into a single shelve.Shelf instance.
    VaultDir creates a file for each object. VaultDirShelf creates only one file for a stored object and everything
    else it references.
    """

    def __init__(self, d=None):
        super().__init__()
        self._d = tempfile.mkdtemp() if d is None else d
        self._dict = None  # will be initialized at each call to store() or load()
        self._lock = threading.Lock()  # protecting access to self._dict

    @contextlib.contextmanager
    def _locked_shelve(self, shelve_path):
        with self._lock:
            self._dict = shelve.open(shelve_path, protocol=-1)  # noqa: SIM115
            yield
            self._dict.close()
            self._dict = None
            self._clear_cache()

    def store(self, o):
        oid = self._get_persistent_id(o) or "TMP-" + str(uuid.uuid4())
        shelve_path = os.path.join(self._d, oid)
        with self._locked_shelve(shelve_path):
            super()._store(o, oid)
        return oid

    def load(self, oid):
        shelve_path = os.path.join(self._d, oid)
        with self._locked_shelve(shelve_path):
            return super().load(oid)

    def keys(self):
        s = set()
        for n in os.listdir(self._d):
            if "." not in n:
                s.add(n)
            else:
                s.add(n[: n.rfind(".")])  # remove the suffix
        return s



from __future__ import annotations
from typing import TYPE_CHECKING
import logging

import networkx

from angr import ailment
from angr.ailment.block_walker import AILBlockViewer
from angr.codenode import BlockNode
from angr.sim_variable import SimMemoryVariable
from angr.knowledge_plugins.functions import Function
from .analysis import Analysis, AnalysesHub

if TYPE_CHECKING:
    from angr.analyses.decompiler.decompiler import Decompiler
    from angr.knowledge_plugins.cfg import CFGModel
    from angr.knowledge_plugins.xrefs import XRefManager

_l = logging.getLogger(name=__name__)


class ProxiNodeTypes:
    """
    Node Type Enums
    """

    Empty = 0
    String = 1
    Function = 2
    FunctionCall = 3
    Integer = 4
    Unknown = 5
    Variable = 6


class BaseProxiNode:
    """
    Base class for all nodes in a proximity graph.
    """

    def __init__(self, type_: int, ref_at: set[int] | None = None):
        self.type_ = type_
        self.ref_at = ref_at

    def __eq__(self, other):
        return isinstance(other, BaseProxiNode) and other.type_ == self.type_ and self.ref_at == other.ref_at

    def __hash__(self):
        return hash((BaseProxiNode, self.type_))


class FunctionProxiNode(BaseProxiNode):
    """
    Proximity node showing current and expanded function calls in graph.
    """

    def __init__(self, func, ref_at: set[int] | None = None):
        super().__init__(ProxiNodeTypes.Function, ref_at=ref_at)
        self.func = func

    def __eq__(self, other):
        return isinstance(other, FunctionProxiNode) and other.type_ == self.type_ and self.func == other.func

    def __hash__(self):
        return hash((FunctionProxiNode, self.func))


class VariableProxiNode(BaseProxiNode):
    """
    Variable arg node
    """

    def __init__(self, addr, name, ref_at: set[int] | None = None):
        super().__init__(ProxiNodeTypes.Variable, ref_at=ref_at)
        self.addr = addr
        self.name = name

    def __eq__(self, other):
        return isinstance(other, VariableProxiNode) and other.type_ == self.type_ and self.addr == other.addr

    def __hash__(self):
        return hash((VariableProxiNode, self.addr))


class StringProxiNode(BaseProxiNode):
    """
    String arg node
    """

    def __init__(self, addr, content, ref_at: set[int] | None = None):
        super().__init__(ProxiNodeTypes.String, ref_at=ref_at)
        self.addr = addr
        self.content = content

    def __eq__(self, other):
        return isinstance(other, StringProxiNode) and other.type_ == self.type_ and self.addr == other.addr

    def __hash__(self):
        return hash((StringProxiNode, self.addr))


class CallProxiNode(BaseProxiNode):
    """
    Call node
    """

    def __init__(self, callee, ref_at: set[int] | None = None, args: tuple[BaseProxiNode] | None = None):
        super().__init__(ProxiNodeTypes.FunctionCall, ref_at=ref_at)
        self.callee = callee
        self.args = args

    def __eq__(self, other):
        return (
            isinstance(other, CallProxiNode)
            and other.type_ == self.type_
            and self.callee == other.callee
            and self.args == other.args
            and self.ref_at == other.ref_at
        )

    def __hash__(self):
        return hash((CallProxiNode, self.callee, self.args))


class IntegerProxiNode(BaseProxiNode):
    """
    Int arg node
    """

    def __init__(self, value: int, ref_at: set[int] | None = None):
        super().__init__(ProxiNodeTypes.Integer, ref_at=ref_at)
        self.value = value

    def __eq__(self, other):
        return isinstance(other, IntegerProxiNode) and self.type_ == other.type_ and self.value == other.value

    def __hash__(self):
        return hash((IntegerProxiNode, self.value))


class UnknownProxiNode(BaseProxiNode):
    """
    Unknown arg node
    """

    def __init__(self, dummy_value: str):
        super().__init__(ProxiNodeTypes.Unknown)
        self.dummy_value = dummy_value

    def __eq__(self, other):
        return (
            isinstance(other, UnknownProxiNode) and self.type_ == other.type_ and self.dummy_value == other.dummy_value
        )

    def __hash__(self):
        return hash((UnknownProxiNode, self.dummy_value))


class ProximityGraphAnalysis(Analysis):
    """
    Generate a proximity graph.
    """

    def __init__(
        self,
        func: Function,
        cfg_model: CFGModel,
        xrefs: XRefManager,
        decompilation: Decompiler | None = None,
        expand_funcs: set[int] | None = None,
    ):
        self._function = func
        self._cfg_model = cfg_model
        self._xrefs = xrefs
        self._decompilation = decompilation
        self._expand_funcs = expand_funcs.copy() if expand_funcs else None

        self.graph: networkx.DiGraph | None = None
        self.handled_stmts = []

        self._work()

    def _condense_blank_nodes(self, graph: networkx.DiGraph) -> None:
        nodes = list(graph.nodes)
        blank_nodes: list[BaseProxiNode] = []

        for node in nodes:
            if isinstance(node, BaseProxiNode) and node.type_ == ProxiNodeTypes.Empty:
                blank_nodes.append(node)
            else:
                if blank_nodes:
                    self._merge_nodes(graph, blank_nodes)
                    blank_nodes = []

        if blank_nodes:
            self._merge_nodes(graph, blank_nodes)

    def _merge_nodes(self, graph: networkx.DiGraph, nodes: list[BaseProxiNode]) -> None:
        for node in nodes:
            predecessors = set(graph.predecessors(node))
            successors = set(graph.successors(node))

            for pred in predecessors:
                for succ in successors:
                    edge_data = graph.get_edge_data(pred, node) or {}
                    graph.add_edge(pred, succ, **edge_data)

            graph.remove_node(node)

    def _work(self):
        self.graph = networkx.DiGraph()

        # initial function
        func_proxi_node = FunctionProxiNode(self._function)

        if not self._decompilation:
            to_expand = self._process_function(self._function, self.graph, func_proxi_node=func_proxi_node)
        else:
            to_expand = self._process_decompilation(
                self.graph, decompilation=self._decompilation, func_proxi_node=func_proxi_node
            )

        for func_node in to_expand:
            if self._expand_funcs:
                self._expand_funcs.discard(func_node.func.addr)

            subgraph = networkx.DiGraph()
            dec = self._decompilation.project.analyses.Decompiler(func_node.func, cfg=self._decompilation._cfg)
            if not dec:
                sub_expand = self._process_function(func_node.func, subgraph, func_proxi_node=func_node)
            else:
                sub_expand = self._process_decompilation(subgraph, decompilation=dec, func_proxi_node=func_node)

            to_expand.extend(sub_expand)

            # merge subgraph into the original graph
            self.graph.add_nodes_from(subgraph.nodes())
            self.graph.add_edges_from(subgraph.edges())

        # condense blank nodes after the graph has been constructed
        self._condense_blank_nodes(self.graph)

    def _endnode_connector(self, func: Function, subgraph: networkx.DiGraph):
        """
        Properly connect expanded function call's to proximity graph.
        """
        current_node = None
        successors = []
        # Get successor node of the current function node
        for node in self.graph.nodes():
            if isinstance(node, FunctionProxiNode) and node.func == func:
                current_node = node
                successors = list(self.graph.succ[node])
                # Remove edge FunctionProxiNode->successors
                for succ in successors:
                    self.graph.remove_edge(node, succ)
                break

        if successors:
            # add edges subgraph_end_nodes->successor
            end_nodes = [n for n in subgraph.nodes() if subgraph.in_degree(n) >= 1 and subgraph.out_degree(n) == 0]

            # handle subgraphs that are empty
            if not end_nodes:
                end_nodes.append(current_node)
            for end_node in end_nodes:
                for succ in successors:
                    subgraph.add_edge(end_node, succ)

    def _process_function(
        self, func: Function, graph: networkx.DiGraph, func_proxi_node: FunctionProxiNode | None = None
    ) -> list[FunctionProxiNode]:
        to_expand: list[FunctionProxiNode] = []
        found_blocks: dict[BlockNode, BaseProxiNode] = {}

        # function calls
        for n_ in func.nodes:
            if isinstance(n_, Function):
                func_node = n_
                for block, _, data in func.transition_graph.in_edges(func_node, data=True):
                    if "ins_addr" in data:
                        if (
                            self._expand_funcs and func_node.addr in self._expand_funcs
                        ):  # pylint:disable=unsupported-membership-test
                            node = FunctionProxiNode(func_node, ref_at={data["ins_addr"]})
                            to_expand.append(node)
                        else:
                            node = CallProxiNode(func_node, ref_at={data["ins_addr"]})
                        found_blocks[block] = node

        # subgraph check - do before in case of recursion
        subgraph = self.graph != graph

        for edge in func.graph.edges:
            nodes = ()
            for block in edge:
                if block in found_blocks:
                    nodes += (found_blocks[block],)
                else:
                    nodes += (BaseProxiNode(ProxiNodeTypes.Empty, {block.addr}),)
            graph.add_edge(*nodes)

        # Append FunctionProxiNode before Graph
        root_node = [n for n, d in graph.in_degree() if d == 0]
        if root_node:
            graph.add_edge(func_proxi_node, root_node[0])

        # Draw edge from subgraph endnodes to current node's successor
        if subgraph:
            self._endnode_connector(func_proxi_node.func, graph)

        return to_expand

    # TODO look into harnessing ailblock_walker to find Strings, Constants, or Function Calls
    def _arg_handler(self, arg, args, string_refs):
        if isinstance(arg, ailment.Expr.Const):
            # is it a reference to a string?
            if arg.value in self._cfg_model.memory_data:
                md = self._cfg_model.memory_data[arg.value]
                if md.sort == "string":
                    # Yes!
                    args.append(StringProxiNode(arg.value, md.content))
                    string_refs.add(arg.value)
            else:
                # not a string. present it as a constant integer
                args.append(IntegerProxiNode(arg.value, None))
        elif isinstance(arg, ailment.expression.Load):
            if arg.variable is not None:
                args.append(VariableProxiNode(arg.variable.addr, arg.variable.name))
            elif arg.addr.variable is not None:
                if isinstance(arg.addr.variable, SimMemoryVariable):
                    args.append(VariableProxiNode(arg.addr.variable.addr, arg.addr.variable.name))
                else:
                    args.append(VariableProxiNode(None, arg.addr.variable.name))
            else:
                args.append(UnknownProxiNode("l_"))
        elif isinstance(arg, ailment.expression.StackBaseOffset):
            if arg.variable is not None:
                args.append(VariableProxiNode(arg.variable, arg.variable.name))
            else:
                args.append(UnknownProxiNode("s_"))
        else:
            args.append(UnknownProxiNode("_"))

    def _process_decompilation(
        self, graph: networkx.DiGraph, decompilation: Decompiler, func_proxi_node: FunctionProxiNode | None = None
    ) -> list[FunctionProxiNode]:
        to_expand: list[FunctionProxiNode] = []

        # dedup
        string_refs: set[int] = set()

        # Walk the clinic structure to dump string references and function calls
        ail_graph = decompilation.clinic.cc_graph

        def _handle_Call(
            stmt_idx: int, stmt: ailment.Stmt.Call, block: ailment.Block | None  # pylint:disable=unused-argument
        ):  # pylint:disable=unused-argument
            if isinstance(stmt.target, ailment.Expr.Const) and self.kb.functions.contains_addr(stmt.target.value):
                func_node = self.kb.functions[stmt.target.value]
                ref_at = {stmt.tags["ins_addr"]}

                # extract arguments
                args = []
                if stmt.args:
                    for arg in stmt.args:
                        self._arg_handler(arg, args, string_refs)

                if (
                    self._expand_funcs and func_node.addr in self._expand_funcs
                ):  # pylint:disable=unsupported-membership-test
                    new_node = FunctionProxiNode(func_node, ref_at=ref_at)
                    if new_node not in to_expand:
                        to_expand.append(new_node)
                else:
                    new_node = CallProxiNode(func_node, ref_at=ref_at, args=tuple(args) if args is not None else None)

                # stmt has been properly handled, add the proxi node to the list
                self.handled_stmts.append(new_node)

        # This should have the same functionality as the previous handler
        def _handle_CallExpr(
            expr_idx: int,
            expr: ailment.Stmt.Call,
            stmt_idx: int,  # pylint:disable=unused-argument
            stmt: ailment.Stmt.Statement,  # pylint:disable=unused-argument
            block: ailment.Block | None,
        ):  # pylint:disable=unused-argument
            _handle_Call(stmt_idx, expr, block)

        # subgraph check - do before in case of recursion
        subgraph = self.graph != graph

        # Keep all default handlers, but overwrite necessary ones:
        bw = AILBlockViewer()
        bw.stmt_handlers[ailment.Stmt.Call] = _handle_Call
        bw.expr_handlers[ailment.Stmt.Call] = _handle_CallExpr

        # Custom Graph walker, go through AIL edges
        for ail_edge in ail_graph.edges:
            new_edge = ()
            # Handle both blocks in edge
            for block in ail_edge:
                bw.walk(block)

                # Check to see if our custom _handle's already dealt with the nodes
                if self.handled_stmts:
                    # Add each handled stmt to the graph in a linear fashion []->[]->[]
                    # Linear is fine since stmts come from the same block and are handled in order of appearance
                    for idx, current in enumerate(self.handled_stmts):
                        # Skip Head of List to avoid out of bounds error
                        if idx > 0:
                            graph.add_edge(self.handled_stmts[idx - 1], current)
                    # If first block in edge, LAST handled node -> next node
                    # Else (second block in edge), prev node -> FIRST handled node
                    proxi_node = self.handled_stmts[-1] if block == ail_edge[0] else self.handled_stmts[0]

                    # Clear the handled stmts for next block
                    self.handled_stmts = []

                # Block stmts weren't handled, make a simple empty proxi node
                else:
                    proxi_node = BaseProxiNode(ProxiNodeTypes.Empty, {block.addr})

                # Create edge from block_1 (proxi_node_1) -> block_2 (proxi_node_2)
                # both blocks may have multiple proxi nodes, but these are already added to the graph
                new_edge += (proxi_node,)

            # Add the new edge for this pair of blocks
            graph.add_edge(*new_edge)

        # Append FunctionProxiNode before Graph
        root_node = [n for n, d in graph.in_degree() if d == 0]
        if root_node:
            graph.add_edge(func_proxi_node, root_node[0])

        # Draw edge from subgraph endnodes to current node's successor
        if subgraph:
            self._endnode_connector(func_proxi_node.func, graph)

        return to_expand


AnalysesHub.register_default("Proximity", ProximityGraphAnalysis)



# pylint:disable=missing-class-docstring
from __future__ import annotations


class AngrError(Exception):
    pass


class AngrRuntimeError(RuntimeError):
    pass


class AngrValueError(AngrError, ValueError):
    pass


class AngrLifterError(AngrError):
    pass


class AngrExitError(AngrError):
    pass


class AngrPathError(AngrError):
    pass


class AngrVaultError(AngrError):
    pass


class PathUnreachableError(AngrPathError):
    pass


class SimulationManagerError(AngrError):
    pass


class AngrInvalidArgumentError(AngrError):
    pass


class AngrSurveyorError(AngrError):
    pass


class AngrAnalysisError(AngrError):
    pass


class AngrBladeError(AngrError):
    pass


class AngrBladeSimProcError(AngrBladeError):
    pass


class AngrAnnotatedCFGError(AngrError):
    pass


class AngrBackwardSlicingError(AngrError):
    pass


class AngrCallableError(AngrSurveyorError):
    pass


class AngrCallableMultistateError(AngrCallableError):
    pass


class AngrSyscallError(AngrError):
    pass


class AngrSimOSError(AngrError):
    pass


class AngrAssemblyError(AngrError):
    pass


class AngrTypeError(AngrError, TypeError):
    pass


class AngrMissingTypeError(AngrTypeError):
    def __init__(self, missing_type: str):
        super().__init__(f"Type {missing_type} is missing")
        self.missing_type = missing_type


# Congruency check failure
class AngrIncongruencyError(AngrAnalysisError):
    pass


#
# ForwardAnalysis errors
#


class AngrForwardAnalysisError(AngrError):
    pass


class AngrSkipJobNotice(AngrForwardAnalysisError):
    pass


class AngrDelayJobNotice(AngrForwardAnalysisError):
    pass


class AngrJobMergingFailureNotice(AngrForwardAnalysisError):
    pass


class AngrJobWideningFailureNotice(AngrForwardAnalysisError):
    pass


#
# CFG errors
#


class AngrCFGError(AngrError):
    pass


#
# VFG Errors and notices
#


class AngrVFGError(AngrError):
    pass


class AngrVFGRestartAnalysisNotice(AngrVFGError):
    pass


#
# Data graph errors
#


class AngrDataGraphError(AngrAnalysisError):
    # TODO: deprecated
    pass


class AngrDDGError(AngrAnalysisError):
    pass


#
# Loop analysis
#


class AngrLoopAnalysisError(AngrAnalysisError):
    pass


#
# Exploration techniques
#


class AngrExplorationTechniqueError(AngrError):
    pass


class AngrExplorerError(AngrExplorationTechniqueError):
    pass


class AngrDirectorError(AngrExplorationTechniqueError):
    pass


class AngrTracerError(AngrExplorationTechniqueError):
    pass


#
# VariableRecovery errors
#


class AngrVariableRecoveryError(AngrAnalysisError):
    pass


#
# AngrDB errors
#


class AngrDBError(AngrError):
    pass


class AngrCorruptDBError(AngrDBError):
    pass


class AngrIncompatibleDBError(AngrDBError):
    pass


#
# Tracer
#


class TracerEnvironmentError(AngrError):
    pass


#
# Simulation errors
#


class SimError(Exception):
    bbl_addr = None
    stmt_idx = None
    ins_addr = None
    executed_instruction_count = None
    guard = None

    def record_state(self, state):
        self.bbl_addr = state.scratch.bbl_addr
        self.stmt_idx = state.scratch.stmt_idx
        self.ins_addr = state.scratch.ins_addr
        self.executed_instruction_count = state.history.recent_instruction_count
        self.guard = state.scratch.guard
        return self


#
# State-related errors
#


class SimStateError(SimError):
    pass


class SimMergeError(SimStateError):
    pass


class SimMemoryError(SimStateError):
    pass


class SimMemoryMissingError(SimMemoryError):
    def __init__(self, missing_addr, missing_size, *args):
        super().__init__(missing_addr, missing_size, *args)
        self.missing_addr = missing_addr
        self.missing_size = missing_size


class SimAbstractMemoryError(SimMemoryError):
    pass


class SimRegionMapError(SimMemoryError):
    pass


class SimMemoryLimitError(SimMemoryError):
    pass


class SimMemoryAddressError(SimMemoryError):
    pass


class SimFastMemoryError(SimMemoryError):
    pass


class SimEventError(SimStateError):
    pass


class SimPosixError(SimStateError):
    pass


class SimFilesystemError(SimError):
    pass


class SimSymbolicFilesystemError(SimFilesystemError):
    pass


class SimFileError(SimMemoryError, SimFilesystemError):
    pass


class SimHeapError(SimStateError):
    pass


#
# Error class during VEX parsing
#


class SimUnsupportedError(SimError):
    pass


#
# Solver-related errors
#


class SimSolverError(SimError):
    pass


class SimSolverModeError(SimSolverError):
    pass


class SimSolverOptionError(SimSolverError):
    pass


class SimValueError(SimSolverError):
    pass


class SimUnsatError(SimValueError):
    pass


#
# SimIROp errors
#


class SimOperationError(SimError):
    pass


class UnsupportedIROpError(SimOperationError, SimUnsupportedError):
    pass


#
# SimIRExpr errors
#


class SimExpressionError(SimError):
    pass


class UnsupportedIRExprError(SimExpressionError, SimUnsupportedError):
    pass


class SimCCallError(SimExpressionError):
    pass


class UnsupportedCCallError(SimCCallError, SimUnsupportedError):
    pass


class SimUninitializedAccessError(SimExpressionError):
    def __init__(self, expr_type, expr):
        SimExpressionError.__init__(self)
        self.expr_type = expr_type
        self.expr = expr

    def __repr__(self):
        return f"SimUninitializedAccessError (expr {self.expr} is used as {self.expr_type})"

    def __reduce__(self):
        return (SimUninitializedAccessError, (self.expr_type, self.expr))


#
# SimIRStmt errors
#


class SimStatementError(SimError):
    pass


class UnsupportedIRStmtError(SimStatementError, SimUnsupportedError):
    pass


class UnsupportedDirtyError(UnsupportedIRStmtError, SimUnsupportedError):
    pass


class SimMissingTempError(SimValueError, IndexError):
    pass


#
# Engine-related errors
#


class SimEngineError(SimError):
    pass


class SimIRSBError(SimEngineError):
    pass


class SimTranslationError(SimEngineError):
    pass


class SimProcedureError(SimEngineError):
    pass


class SimProcedureArgumentError(SimProcedureError):
    pass


class SimShadowStackError(SimProcedureError):
    pass


class SimFastPathError(SimEngineError):
    pass


class SimIRSBNoDecodeError(SimIRSBError):
    pass


class AngrUnsupportedSyscallError(AngrSyscallError, SimProcedureError, SimUnsupportedError):
    pass


UnsupportedSyscallError = AngrUnsupportedSyscallError


class SimReliftException(SimEngineError):
    def __init__(self, state):
        super().__init__()
        self.state = state


#
# SimSlicer errors
#


class SimSlicerError(SimError):
    pass


#
# SimAction errors
#


class SimActionError(SimError):
    pass


#
# SimCC errors
#


class SimCCError(SimError):
    pass


#
# UCManager errors
#


class SimUCManagerError(SimError):
    pass


class SimUCManagerAllocationError(SimUCManagerError):
    pass


#
# SimUnicorn errors
#


class SimUnicornUnsupport(SimError):
    pass


class SimUnicornError(SimError):
    pass


class SimUnicornSymbolic(SimError):
    pass


#
# Call-stack Errors
#


class SimEmptyCallStackError(SimError):
    pass


#
# SimStateOptions Errors
#


class SimStateOptionsError(SimError):
    pass


#
# Errors that may be handled by exception handling
#


class SimException(SimError):
    pass


class SimSegfaultException(SimException, SimMemoryError):
    def __init__(self, addr, reason, original_addr=None):
        self.addr = addr
        self.reason = reason
        self.original_addr = original_addr
        super(SimSegfaultError, self).__init__(f"{addr:#x} ({reason})")

    def __repr__(self):
        return "SimSegfaultException({:#x} ({}{})".format(
            self.addr,
            self.reason,
            (f", original {self.original_addr.__repr__(max_depth=3)}") if self.original_addr is not None else "",
        )

    def __reduce__(self):
        return (SimSegfaultException, (self.addr, self.reason, self.original_addr))


SimSegfaultError = SimSegfaultException


class SimZeroDivisionException(SimException, SimOperationError):
    pass


class AngrNoPluginError(AngrError):
    pass


#
# Concrete Targets Execution errors
#


class SimConcreteMemoryError(AngrError):
    pass


class SimConcreteRegisterError(AngrError):
    pass


class SimConcreteBreakpointError(AngrError):
    pass


#
# Decompiler errors
#


class AngrDecompilationError(AngrError):
    pass


class UnsupportedNodeTypeError(AngrError, NotImplementedError):
    pass



from __future__ import annotations
import logging

from angr.procedures.stubs.format_parser import ScanfFormatParser

l = logging.getLogger(name=__name__)


class sscanf(ScanfFormatParser):
    # pylint:disable=arguments-differ,unused-argument
    def run(self, data, fmt):
        fmt_str = self._parse(fmt)
        return fmt_str.interpret(self.va_arg, addr=data)



from __future__ import annotations
import logging

from pypcode import OpCode, Varnode, PcodeOp
import pypcode

from angr.utils.constants import DEFAULT_STATEMENT
from angr.engines.pcode.lifter import IRSB
from .block import Block
from .statement import Statement, Assignment, Store, Jump, ConditionalJump, Return, Call
from .expression import Expression, DirtyExpression, Const, Register, Tmp, UnaryOp, BinaryOp, Load, Convert

# FIXME: Convert, ITE
from .manager import Manager
from .converter_common import Converter


log = logging.getLogger(name=__name__)

# FIXME: Not all ops are mapped to AIL expressions!
opcode_to_generic_name = {
    # OpCode.MULTIEQUAL        : '',
    # OpCode.INDIRECT          : '',
    # OpCode.PIECE             : '',
    # OpCode.SUBPIECE          : '',
    OpCode.INT_EQUAL: "CmpEQ",
    OpCode.INT_NOTEQUAL: "CmpNE",
    OpCode.INT_SLESS: "CmpLTs",
    OpCode.INT_SLESSEQUAL: "CmpLEs",
    OpCode.INT_LESS: "CmpLT",
    OpCode.INT_LESSEQUAL: "CmpLE",
    # OpCode.INT_ZEXT          : '',
    # OpCode.INT_SEXT          : '',
    OpCode.INT_ADD: "Add",
    OpCode.INT_SUB: "Sub",
    OpCode.INT_CARRY: "Carry",
    OpCode.INT_SCARRY: "SCarry",
    OpCode.INT_SBORROW: "SBorrow",
    # OpCode.INT_2COMP         : '',
    OpCode.INT_NEGATE: "Neg",
    OpCode.INT_XOR: "Xor",
    OpCode.INT_AND: "And",
    OpCode.INT_OR: "Or",
    OpCode.INT_LEFT: "Shl",
    OpCode.INT_RIGHT: "Shr",
    OpCode.INT_SRIGHT: "Sar",
    OpCode.INT_MULT: "Mul",
    OpCode.INT_DIV: "Div",
    # OpCode.INT_SDIV          : '',
    # OpCode.INT_REM           : '',
    # OpCode.INT_SREM          : '',
    OpCode.BOOL_NEGATE: "Not",
    OpCode.BOOL_XOR: "Xor",
    OpCode.BOOL_AND: "LogicalAnd",
    OpCode.BOOL_OR: "LogicalOr",
    # OpCode.CAST              : '',
    # OpCode.PTRADD            : '',
    # OpCode.PTRSUB            : '',
    # OpCode.FLOAT_EQUAL       : '',
    # OpCode.FLOAT_NOTEQUAL    : '',
    # OpCode.FLOAT_LESS        : '',
    # OpCode.FLOAT_LESSEQUAL   : '',
    # OpCode.FLOAT_NAN         : '',
    # OpCode.FLOAT_ADD         : '',
    OpCode.FLOAT_DIV: "Div",
    OpCode.FLOAT_MULT: "Mul",
    OpCode.FLOAT_SUB: "Sub",
    # OpCode.FLOAT_NEG         : '',
    # OpCode.FLOAT_ABS         : '',
    # OpCode.FLOAT_SQRT        : '',
    # OpCode.FLOAT_INT2FLOAT   : '',
    # OpCode.FLOAT_FLOAT2FLOAT : '',
    # OpCode.FLOAT_TRUNC       : '',
    # OpCode.FLOAT_CEIL        : '',
    # OpCode.FLOAT_FLOOR       : '',
    # OpCode.FLOAT_ROUND       : '',
    # OpCode.SEGMENTOP         : '',
    # OpCode.CPOOLREF          : '',
    # OpCode.NEW               : '',
    # OpCode.INSERT            : '',
    # OpCode.EXTRACT           : '',
    # OpCode.POPCOUNT          : '',
}


class PCodeIRSBConverter(Converter):
    """
    Converts a p-code IRSB to an AIL block
    """

    _current_op: PcodeOp

    @staticmethod
    def convert(irsb: IRSB, manager: Manager):  # pylint:disable=arguments-differ
        """
        Convert the given IRSB to an AIL block

        :param irsb:    IRSB to convert
        :param manager: Manager to use
        :return:        Converted block
        """
        return PCodeIRSBConverter(irsb, manager)._convert()

    def __init__(self, irsb: IRSB, manager: Manager):
        self._irsb = irsb
        self._manager = manager
        self._statements = []
        self._next_ins_addr = None
        self._current_behavior = None
        self._statement_idx = 0

        # Remap all uniques s.t. they are write-once with values starting from 0
        self._unique_tracker: dict[int, tuple[int, int]] = {}
        self._unique_counter = 0

        self._special_op_handlers = {
            OpCode.COPY: self._convert_copy,
            OpCode.INT_ZEXT: self._convert_zext,
            OpCode.INT_SEXT: self._convert_sext,
            OpCode.LOAD: self._convert_load,
            OpCode.STORE: self._convert_store,
            OpCode.BRANCH: self._convert_branch,
            OpCode.CBRANCH: self._convert_cbranch,
            OpCode.BRANCHIND: self._convert_branchind,
            OpCode.CALL: self._convert_call,
            OpCode.CALLIND: self._convert_callind,
            OpCode.CALLOTHER: self._convert_callother,
            OpCode.RETURN: self._convert_ret,
            OpCode.MULTIEQUAL: self._convert_multiequal,
            OpCode.INDIRECT: self._convert_indirect,
            OpCode.SEGMENTOP: self._convert_segment_op,
            OpCode.CPOOLREF: self._convert_cpool_ref,
            OpCode.NEW: self._convert_new,
            OpCode.FLOAT_INT2FLOAT: self._convert_int2float,
            OpCode.FLOAT_FLOAT2FLOAT: self._convert_float2float,
        }

        manager.tyenv = None
        manager.block_addr = irsb.addr
        manager.vex_stmt_idx = DEFAULT_STATEMENT  # Reset after loop. Necessary?

    def _convert(self) -> Block:
        """
        Convert the given IRSB to an AIL Block
        """
        self._statement_idx = 0

        for op in self._irsb._ops:
            self._current_op = op
            if op.opcode == pypcode.OpCode.IMARK:
                self._manager.ins_addr = op.inputs[0].offset
                self._next_ins_addr = op.inputs[-1].offset + op.inputs[-1].size
            else:
                assert self._irsb.behaviors is not None
                self._current_behavior = self._irsb.behaviors.get_behavior_for_opcode(self._current_op.opcode)
                self._convert_current_op()
            self._statement_idx += 1

            if (
                "sparc:" in self._irsb.arch.name
                and self._irsb.arch.bits == 32
                and self._current_op.opcode == OpCode.CALL
            ):
                break

        return Block(self._irsb.addr, self._irsb.size, statements=self._statements)

    def _convert_current_op(self) -> None:
        """
        Convert the current op to corresponding AIL statement
        """
        assert self._current_behavior is not None

        is_special = self._current_behavior.opcode in self._special_op_handlers

        if is_special:
            try:
                self._special_op_handlers[self._current_behavior.opcode]()
            except NotImplementedError as ex:
                log.warning("Unsupported opcode: %s", ex)
        elif self._current_behavior.is_unary:
            self._convert_unary()
        else:
            self._convert_binary()

    def _convert_unary(self) -> None:
        """
        Convert the current unary op to corresponding AIL statement
        """
        opcode = self._current_op.opcode

        op = opcode_to_generic_name.get(opcode)
        in1 = self._get_value(self._current_op.inputs[0])
        if op is None:
            log.warning("p-code: Unsupported opcode of type %s", opcode.__name__)
            out = DirtyExpression(
                self._manager.next_atom(),
                opcode.__name__,
                [],
                bits=self._current_op.output.size * 8 if self._current_op.output is not None else 1,
            )
        else:
            out = UnaryOp(self._manager.next_atom(), op, in1, ins_addr=self._manager.ins_addr)

        stmt = self._set_value(self._current_op.output, out)
        self._statements.append(stmt)

    def _convert_binary(self) -> None:
        """
        Convert the current binary op to corresponding AIL statement
        """
        opcode = self._current_op.opcode
        op = opcode_to_generic_name.get(opcode)
        in1 = self._get_value(self._current_op.inputs[0])
        in2 = self._get_value(self._current_op.inputs[1])
        if op is None:
            log.warning("p-code: Unsupported opcode of type %s.", opcode.__name__)
            out = DirtyExpression(
                self._manager.next_atom(),
                opcode.__name__,
                [],
                bits=self._current_op.output.size * 8 if self._current_op.output is not None else 1,
            )
        else:
            # fix op name for signed comparisons
            signed = op.startswith("Cmp") and op.endswith("s")
            if signed and op.endswith("s"):
                op = op[:-1]
            out = BinaryOp(self._manager.next_atom(), op, [in1, in2], signed, ins_addr=self._manager.ins_addr)

        # Zero-extend 1-bit results
        zextend_ops = {
            OpCode.INT_EQUAL,
            OpCode.INT_NOTEQUAL,
            OpCode.INT_SLESS,
            OpCode.INT_SLESSEQUAL,
            OpCode.INT_LESS,
            OpCode.INT_LESSEQUAL,
        }
        if opcode in zextend_ops:
            out = Convert(self._manager.next_atom(), 1, self._current_op.output.size * 8, False, out)

        stmt = self._set_value(self._current_op.output, out)
        self._statements.append(stmt)

    def _map_register_name(self, varnode: Varnode) -> int:
        """
        Map SLEIGH register offset to ArchInfo register offset based on name.

        :param varnode: The varnode to translate
        :return:        The register file offset
        """
        # FIXME: Will need performance optimization
        # FIXME: Should not get trans object this way. Moreover, should have a
        #        faster mapping method than going through trans
        reg_name = varnode.getRegisterName()
        try:
            reg_offset = self._manager.arch.get_register_offset(reg_name.lower())
            log.debug("Mapped register '%s' to offset %x", reg_name, reg_offset)
        except ValueError:
            reg_offset = varnode.offset + 0x100000
            log.warning("Could not map register '%s' from archinfo. Mapping to %x", reg_name, reg_offset)
        return reg_offset

    def _remap_temp(self, offset: int, size: int, is_write: bool) -> int | None:
        """
        Remap any unique space addresses such that they are written only once

        :param offset:   The unique space address
        :param is_write: Whether the access is a write or a read
        :return:         The remapped temporary register index
        """
        if is_write:
            self._unique_tracker[offset] = self._unique_counter, size
            self._unique_counter += 1
            return self._unique_tracker[offset][0]
        if offset in self._unique_tracker:
            return self._unique_tracker[offset][0]
        # this might be a partial access of an existing temporary variable. return None for now
        return None

    def _convert_varnode(self, varnode: Varnode, is_write: bool) -> Expression:
        """
        Convert a varnode to a corresponding AIL expression

        :param varnode:  The varnode to remap
        :param is_write: Whether the varnode is being read or written to
        :return:         The corresponding AIL expression
        """
        space_name = varnode.space.name
        size = varnode.size * 8

        if space_name == "const":
            return Const(self._manager.next_atom(), None, varnode.offset, size)
        if space_name == "register":
            offset = self._map_register_name(varnode)
            return Register(
                self._manager.next_atom(),
                None,
                offset,
                size,
                reg_name=varnode.getRegisterName(),
                ins_addr=self._manager.ins_addr,
            )
        if space_name == "unique":
            offset = self._remap_temp(varnode.offset, varnode.size, is_write)
            if offset is None:
                # this might be a partial access of an existing temporary variable
                unique_offset = None
                for delta in range(-1, -8, -1):
                    if varnode.offset + delta in self._unique_tracker:
                        unique_offset = varnode.offset + delta
                        break
                assert unique_offset is not None, "Cannot find the source unique variable"
                # TODO: Check size
                _, ori_tmp_size = self._unique_tracker[unique_offset]
                t = Tmp(self._manager.next_atom(), None, unique_offset, ori_tmp_size * 8)
                # FIXME: Asserting BE
                right_shift_amount = varnode.offset + varnode.size - (unique_offset + ori_tmp_size)
                if right_shift_amount != 0:
                    t = BinaryOp(
                        self._manager.next_atom(),
                        "Shr",
                        [t, Const(self._manager.next_atom(), None, right_shift_amount * 8, 8)],
                        False,
                        ins_addr=self._manager.ins_addr,
                    )
                return Convert(self._manager.next_atom(), t.bits, size, False, t, ins_addr=self._manager.ins_addr)

            return Tmp(self._manager.next_atom(), None, offset, size)
        if space_name in ["ram", "mem"]:
            assert not is_write
            addr = Const(self._manager.next_atom(), None, varnode.offset, self._manager.arch.bits)
            # Note: Load takes bytes, not bits, for size
            return Load(
                self._manager.next_atom(),
                addr,
                varnode.size,
                self._manager.arch.memory_endness,
                ins_addr=self._manager.ins_addr,
            )
        raise NotImplementedError

    def _set_value(self, varnode: Varnode, value: Expression) -> Statement:
        """
        Create the appropriate assignment statement to store to a varnode

        This method stores to the appropriate register, or unique space,
        depending on the space indicated by the varnode.

        :param varnode: Varnode to store into
        :param value:   Value to store
        :return:        Corresponding AIL statement
        """
        space_name = varnode.space.name

        if space_name in ["register", "unique"]:
            return Assignment(
                self._statement_idx, self._convert_varnode(varnode, True), value, ins_addr=self._manager.ins_addr
            )
        if space_name in ["ram", "mem"]:
            addr = Const(self._manager.next_atom(), None, varnode.offset, self._manager.arch.bits)
            return Store(
                self._statement_idx,
                addr,
                value,
                varnode.size,
                self._manager.arch.memory_endness,
                ins_addr=self._manager.ins_addr,
            )
        raise NotImplementedError

    def _get_value(self, varnode: Varnode) -> Expression:
        """
        Create the appropriate expression to load from a varnode

        This method loads from the appropriate const, register, unique, or RAM
        space, depending on the space indicated by the varnode.

        :param varnode: Varnode to load from.
        :return:        Value loaded
        """
        return self._convert_varnode(varnode, False)

    def _convert_copy(self) -> None:
        """
        Convert copy operation
        """
        out = self._current_op.output
        inp = self._get_value(self._current_op.inputs[0])
        stmt = self._set_value(out, inp)
        self._statements.append(stmt)

    def _convert_zext(self) -> None:
        """
        Convert zext operation
        """
        out = self._current_op.output
        inp = Convert(
            self._manager.next_atom(),
            self._current_op.inputs[0].size * 8,
            out.size * 8,
            False,
            self._get_value(self._current_op.inputs[0]),
        )
        stmt = self._set_value(out, inp)
        self._statements.append(stmt)

    def _convert_sext(self) -> None:
        """
        Convert the signed extension operation
        """
        out = self._current_op.output
        inp = Convert(
            self._manager.next_atom(),
            self._current_op.inputs[0].size * 8,
            out.size * 8,
            False,
            self._get_value(self._current_op.inputs[0]),
        )
        stmt = self._set_value(out, inp)
        self._statements.append(stmt)

    def _convert_negate(self) -> None:
        """
        Convert bool negate operation
        """
        out = self._current_op.output
        inp = self._get_value(self._current_op.inputs[0])

        cval = Const(self._manager.next_atom(), None, 0, self._current_op.inputs[0].size * 8)

        expr = BinaryOp(self._manager.next_atom(), "CmpEQ", [inp, cval], signed=False, ins_addr=self._manager.ins_addr)

        stmt = self._set_value(out, expr)
        self._statements.append(stmt)

    def _convert_load(self) -> None:
        """
        Convert a p-code load operation
        """
        spc = self._current_op.inputs[0].getSpaceFromConst()
        out = self._current_op.output
        spc_name = spc.name.lower()
        assert spc_name in {"ram", "mem", "register"}
        if spc_name == "register":
            # load from register
            res = self._get_value(self._current_op.inputs[1])
            stmt = self._set_value(out, res)
        else:
            # load from memory
            off = self._get_value(self._current_op.inputs[1])
            res = Load(
                self._manager.next_atom(),
                off,
                self._current_op.output.size,
                self._manager.arch.memory_endness,
                ins_addr=self._manager.ins_addr,
            )
            stmt = self._set_value(out, res)
        self._statements.append(stmt)

    def _convert_store(self) -> None:
        """
        Convert a p-code store operation
        """
        spc = self._current_op.inputs[0].getSpaceFromConst()
        spc_name = spc.name.lower()
        assert spc_name in {"ram", "mem", "register"}
        if spc_name == "register":
            # store to register
            out = self._current_op.inputs[2]
            res = self._get_value(self._current_op.inputs[1])
            stmt = self._set_value(out, res)
        else:
            # store to memory
            off = self._get_value(self._current_op.inputs[1])
            data = self._get_value(self._current_op.inputs[2])
            log.debug("Storing %s at offset %s", data, off)
            # self.state.memory.store(off, data, endness=self.project.arch.memory_endness)
            stmt = Store(
                self._statement_idx,
                off,
                data,
                self._current_op.inputs[2].size,
                self._manager.arch.memory_endness,
                ins_addr=self._manager.ins_addr,
            )
        self._statements.append(stmt)

    def _convert_branch(self) -> None:
        """
        Convert a p-code branch operation
        """
        if self._current_op.inputs[0].space == "const":
            raise NotImplementedError("p-code relative branch not supported yet")
        dest_addr = self._current_op.inputs[0].offset

        # special handling: if the previous statement is a ConditionalJump with a None destination address, then we
        # back-patch the previous statement
        dest = Const(self._manager.next_atom(), None, dest_addr, self._manager.arch.bits)
        if self._statements:
            last_stmt = self._statements[-1]
            if isinstance(last_stmt, ConditionalJump) and last_stmt.false_target is None:
                last_stmt.false_target = dest
                return

        stmt = Jump(self._statement_idx, dest, ins_addr=self._manager.ins_addr)
        self._statements.append(stmt)

    def _convert_cbranch(self) -> None:
        """
        Convert a p-code conditional branch operation
        """
        if self._current_op.inputs[0].space == "const":
            raise NotImplementedError("p-code relative branch not supported yet")
        dest_addr = self._current_op.inputs[0].offset
        cond = self._get_value(self._current_op.inputs[1])
        cval = Const(self._manager.next_atom(), None, 0, cond.bits)
        condition = BinaryOp(self._manager.next_atom(), "CmpNE", [cond, cval], signed=False)
        dest = Const(self._manager.next_atom(), None, dest_addr, self._manager.arch.bits)
        if self._irsb._ops[-1] is self._current_op:
            # if the cbranch op is the last op, then we need to generate a fallthru target
            fallthru = Const(
                self._manager.next_atom(),
                None,
                self._next_ins_addr,
                self._manager.arch.bits,
            )
        else:
            # there will be a Jump statement that follows the cbranch
            fallthru = None
        stmt = ConditionalJump(self._statement_idx, condition, dest, fallthru, ins_addr=self._manager.ins_addr)
        self._statements.append(stmt)

    def _convert_ret(self) -> None:
        """
        Convert a p-code return operation
        """
        stmt = Return(
            self._statement_idx,
            [],
            ins_addr=self._manager.ins_addr,
            vex_block_addr=self._manager.block_addr,
            vex_stmt_idx=DEFAULT_STATEMENT,
        )
        self._statements.append(stmt)

    def _convert_branchind(self) -> None:
        """
        Convert a p-code indirect branch operation
        """
        dest = self._get_value(self._current_op.inputs[0])
        stmt = Jump(self._statement_idx, dest, ins_addr=self._manager.ins_addr)
        self._statements.append(stmt)

    def _convert_call(self) -> None:
        """
        Convert a p-code call operation
        """
        ret_reg_offset = self._manager.arch.ret_offset
        ret_expr = (
            None
            if ret_reg_offset is None
            else Register(None, None, ret_reg_offset, self._manager.arch.bits, ins_addr=self._manager.ins_addr)
        )  # ???
        if self._irsb.next is not None:
            dest = Const(self._manager.next_atom(), None, self._irsb.next.con.value, self._manager.arch.bits)
        else:
            dest = None
        stmt = Call(
            self._manager.next_atom(),
            dest,
            ret_expr=ret_expr,
            ins_addr=self._manager.ins_addr,
            vex_block_addr=self._manager.block_addr,
            vex_stmt_idx=DEFAULT_STATEMENT,
        )
        self._statements.append(stmt)

    def _convert_callind(self) -> None:
        """
        Convert a p-code indirect call operation
        """
        ret_reg_offset = self._manager.arch.ret_offset
        ret_expr = Register(None, None, ret_reg_offset, self._manager.arch.bits, ins_addr=self._manager.ins_addr)  # ???
        dest = self._get_value(self._current_op.inputs[0])
        stmt = Call(
            self._manager.next_atom(),
            dest,
            ret_expr=ret_expr,
            ins_addr=self._manager.ins_addr,
            vex_block_addr=self._manager.block_addr,
            vex_stmt_idx=DEFAULT_STATEMENT,
        )
        self._statements.append(stmt)

    def _convert_int2float(self) -> None:
        """
        Convert INT2FLOAT operation.
        """
        out = self._current_op.output
        inp = Convert(
            self._manager.next_atom(),
            self._current_op.inputs[0].size * 8,
            out.size * 8,
            True,
            self._get_value(self._current_op.inputs[0]),
            from_type=Convert.TYPE_INT,
            to_type=Convert.TYPE_FP,
        )
        stmt = self._set_value(out, inp)
        self._statements.append(stmt)

    def _convert_float2float(self) -> None:
        """
        Convert FLOAT2FLOAT operation.
        """
        out = self._current_op.output
        inp = Convert(
            self._manager.next_atom(),
            self._current_op.inputs[0].size * 8,
            out.size * 8,
            True,
            self._get_value(self._current_op.inputs[0]),
            from_type=Convert.TYPE_FP,
            to_type=Convert.TYPE_FP,
        )
        stmt = self._set_value(out, inp)
        self._statements.append(stmt)

    def _convert_callother(self) -> None:
        raise NotImplementedError("CALLOTHER emulation not currently supported")

    def _convert_multiequal(self) -> None:
        raise NotImplementedError("MULTIEQUAL appearing in unheritaged code?")

    def _convert_indirect(self) -> None:
        raise NotImplementedError("INDIRECT appearing in unheritaged code?")

    def _convert_segment_op(self) -> None:
        raise NotImplementedError("SEGMENTOP emulation not currently supported")

    def _convert_cpool_ref(self) -> None:
        raise NotImplementedError("Cannot currently emulate cpool operator")

    def _convert_new(self) -> None:
        raise NotImplementedError("Cannot currently emulate new operator")



# pylint:disable=no-self-use,too-many-boolean-expressions
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ASubAShrConstShrConst(PeepholeOptimizationExprBase):
    """
    Convert `cdq; sub eax, edx; sar eax, 1` to `eax /= 2`.
    """

    __slots__ = ()

    NAME = "(a - (a >> 31)) >> N => a / 2 ** N (signed)"
    expr_classes = (BinaryOp,)

    def optimize(self, expr: BinaryOp, **kwargs):
        if (
            expr.op == "Sar"
            and len(expr.operands) == 2
            and isinstance(expr.operands[1], Const)
            and expr.operands[1].is_int
            and isinstance(expr.operands[0], BinaryOp)
            and expr.operands[0].op == "Sub"
        ):
            a0, a1 = expr.operands[0].operands
            if (
                isinstance(a1, BinaryOp)
                and a1.op == "Sar"
                and isinstance(a1.operands[1], Const)
                and a1.operands[1].value == 31
                and a0.likes(a1.operands[0])
            ):
                dividend = 2 ** expr.operands[1].value
                return BinaryOp(a0.idx, "Div", [a0, Const(None, None, dividend, expr.bits)], True, **expr.tags)
        return None



# pylint:disable=abstract-method,arguments-differ,assignment-from-no-return
from __future__ import annotations
import logging
from typing import Any
from collections.abc import Callable

from angr.storage.memory_mixins.memory_mixin import MemoryMixin
from angr.utils.dynamic_dictlist import DynamicDictList
from angr.storage.memory_object import SimMemoryObject, SimLabeledMemoryObject
from .base import PageBase
from .cooperation import MemoryObjectSetMixin


l = logging.getLogger(name=__name__)

_MOTYPE = SimMemoryObject | SimLabeledMemoryObject


class MVListPage(
    MemoryObjectSetMixin,
    PageBase,
):
    """
    MVListPage allows storing multiple values at the same location.

    Each store() may take a value or multiple values.
    Each load() returns an iterator of all values stored at that location.
    """

    def __init__(self, memory=None, content=None, sinkhole=None, mo_cmp=None, **kwargs):
        super().__init__(**kwargs)

        self.content: DynamicDictList[_MOTYPE | set[_MOTYPE] | None] = (
            DynamicDictList(max_size=memory.page_size, content=content) if content is not None else None
        )
        self.stored_offset = set()
        self._mo_cmp: Callable | None = mo_cmp

        if self.content is None and memory is not None:
            self.content: DynamicDictList[_MOTYPE | set[_MOTYPE] | None] = DynamicDictList(max_size=memory.page_size)

        self.sinkhole: _MOTYPE | None = sinkhole

    def copy(self, memo) -> MVListPage:
        o = super().copy(memo)
        o.content = DynamicDictList(max_size=self.content.max_size, content=self.content)
        o.sinkhole = self.sinkhole
        o.stored_offset = self.stored_offset.copy()
        o._mo_cmp = self._mo_cmp
        return o

    def load(
        self, addr, size=None, endness=None, page_addr=None, memory=None, cooperate=False, **kwargs
    ) -> list[tuple[int, _MOTYPE]]:
        result = []
        last_seen = ...  # ;)

        # loop over the loading range. accumulate a result for each byte, but collapse results from adjacent bytes
        # using the same memory object
        for subaddr in range(addr, addr + size):
            items = self.content[subaddr]
            if items is None:
                items = {self.sinkhole} if self.sinkhole is not None else None
            if items != last_seen:
                if last_seen is None:
                    self._fill(result, subaddr, page_addr, endness, memory, **kwargs)
                result.append((subaddr + page_addr, items))
                last_seen = items

        if last_seen is None:
            self._fill(result, addr + size, page_addr, endness, memory, **kwargs)

        if not cooperate:
            result = self._force_load_cooperation(result, size, endness, memory=memory, **kwargs)
        return result

    def _fill(self, result, addr, page_addr, endness, memory, **kwargs):
        """
        Small utility function for behavior which is duplicated in load

        mutates result to generate a new memory object and replace the last entry in it, which is None. Then, it will
        insert the new memory object into self.content.
        """
        global_end_addr = addr + page_addr
        global_start_addr = result[-1][0]
        size = global_end_addr - global_start_addr
        new_ast = self._default_value(
            global_start_addr,
            size,
            name=f"{memory.id}_{global_start_addr:x}",
            key=(self.category, global_start_addr),
            memory=memory,
            **kwargs,
        )
        new_item = SimMemoryObject(
            new_ast,
            global_start_addr,
            endness=endness,
            byte_width=memory.state.arch.byte_width if memory is not None else 8,
        )
        subaddr_start = global_start_addr - page_addr
        for subaddr in range(subaddr_start, addr):
            self.content[subaddr] = {new_item}
            self.stored_offset.add(subaddr)
        result[-1] = (global_start_addr, new_item)

    def store(self, addr, data, size=None, endness=None, memory=None, cooperate=False, **kwargs):
        super().store(addr, data, size=size, endness=endness, memory=memory, cooperate=cooperate, **kwargs)

        if not cooperate:
            data = self._force_store_cooperation(addr, data, size, endness, memory=memory, **kwargs)

        data: set[_MOTYPE]

        if size == len(self.content) and addr == 0 and len(data) == 1:
            self.sinkhole = next(iter(data))
            self.content = DynamicDictList(max_size=len(self.content))
            self.stored_offset = set()
        else:
            if len(data) == 1:
                # unpack
                data: _MOTYPE = next(iter(data))
            for subaddr in range(addr, addr + size):
                self.content[subaddr] = data
                self.stored_offset.add(subaddr)

    def erase(self, addr, size=None, **kwargs) -> None:
        for off in range(size):
            self.content[addr + off] = None

    def merge(
        self,
        others: list[MVListPage],
        merge_conditions,
        common_ancestor=None,
        *,
        page_addr: int,
        memory: MemoryMixin,
        changed_offsets: set[int] | None = None,
    ):
        if changed_offsets is None:
            changed_offsets = set()
            for other in others:
                changed_offsets |= self.changed_bytes(other, page_addr)

        all_pages: list[MVListPage] = [self, *others]
        if merge_conditions is None:
            merge_conditions = [None] * len(all_pages)

        merged_to = None
        merged_objects = set()
        merged_offsets = set()
        for b in sorted(changed_offsets):
            if merged_to is not None and not b >= merged_to:
                l.info("merged_to = %d ... already merged byte 0x%x", merged_to, b)
                continue
            l.debug("... on byte 0x%x", b)

            memory_object_sets: set[tuple[frozenset[SimMemoryObject], Any]] = set()
            unconstrained_in = []

            # first get a list of all memory objects at that location, and all memories that don't have those bytes
            self_has_memory_object_set = False
            for sm, fv in zip(all_pages, merge_conditions):
                if sm._contains(b, page_addr):
                    l.info("... present in %s", fv)
                    memory_objects = set()
                    for mo in sm.content_gen(b):
                        if mo.includes(page_addr + b):
                            memory_objects.add(mo)
                    memory_object_sets.add((frozenset(memory_objects), fv))
                    if sm is self:
                        self_has_memory_object_set = True
                else:
                    l.info("... not present in %s", fv)
                    unconstrained_in.append((sm, fv))

            if not memory_object_sets:
                continue
            if self_has_memory_object_set and len(memory_object_sets) == 1:
                continue

            mo_sets = {mo_set for mo_set, _ in memory_object_sets}
            mo_bases = set()
            mo_lengths = set()
            endnesses = set()
            for mo_set in mo_sets:
                for mo in mo_set:
                    mo_bases.add(mo.base)
                    mo_lengths.add(mo.length)
                    endnesses.add(mo.endness)

            if not unconstrained_in and not (mo_sets - merged_objects):  # pylint:disable=superfluous-parens
                continue

            # first, optimize the case where we are dealing with the same-sized memory objects
            if len(mo_bases) == 1 and len(mo_lengths) == 1 and not unconstrained_in and len(endnesses) == 1:
                if len(memory_object_sets) == 1:
                    # nothing to merge!
                    continue

                the_endness = next(iter(endnesses))
                to_merge = []
                for mo_set, fv in memory_object_sets:
                    for mo in mo_set:
                        to_merge.append((mo.object, fv))

                # Update `merged_to`
                mo_base = next(iter(mo_bases))
                mo_length = next(iter(mo_lengths))
                size = min(mo_length - (page_addr + b - mo_base), len(self.content) - b)
                merged_to = b + size

                merged_val = self._merge_values(to_merge, mo_length, memory=memory)
                if merged_val is None:
                    # merge_values() determines that we should not attempt to merge this value
                    continue

                # do the replacement
                # TODO: Implement in-place replacement instead of calling store()
                # new_object = self._replace_memory_object(our_mo, merged_val, page_addr, memory.page_size)

                new_mos = {SimMemoryObject(v, mo_base, endness=the_endness) for v in merged_val}
                self.store(b, new_mos, size=size, cooperate=True)

                merged_offsets.add(b)

            else:
                # get the size that we can merge easily. This is the minimum of the size of all memory objects and
                # unallocated spaces.
                min_size = len(self.content) - b
                mask = (1 << memory.state.arch.bits) - 1
                for mo_set in mo_sets:
                    for mo in mo_set:
                        min_size = min(min_size, mo.length - ((b + page_addr - mo.base) & mask))
                for um, _ in unconstrained_in:
                    for i in range(min_size):
                        if um._contains(b + i, page_addr):
                            min_size = i
                            break
                merged_to = b + min_size
                l.info("... determined minimum size of %d", min_size)

                # Now, we have the minimum size. We'll extract/create expressions of that
                # size and merge them
                extracted = []
                if min_size != 0:
                    for mo_set, fv in memory_object_sets:
                        for mo in mo_set:
                            extracted.append((mo.bytes_at(page_addr + b, min_size), fv))
                if not memory.skip_missing_values_during_merging:
                    created = [
                        (self._default_value(None, min_size, name=f"merge_uc_{uc.id}_{b:x}", memory=memory), fv)
                        for uc, fv in unconstrained_in
                    ]
                    to_merge = extracted + created
                else:
                    to_merge = extracted

                merged_val = self._merge_values(to_merge, min_size, memory=memory)
                if merged_val is None:
                    continue

                new_mos = {SimMemoryObject(v, page_addr + b, endness="Iend_BE") for v in merged_val}
                self.store(b, new_mos, size=min_size, cooperate=True)
                merged_offsets.add(b)

        self.stored_offset |= merged_offsets
        return merged_offsets

    def compare(
        self, other: MVListPage, page_addr: int | None = None, memory=None, changed_offsets=None
    ) -> bool:  # pylint: disable=unused-argument
        compared_to = None
        for b in sorted(changed_offsets):
            if compared_to is not None and not b >= compared_to:
                continue

            unconstrained_in = []
            self_has_memory_object_set = False
            memory_object_sets: set[frozenset[SimMemoryObject]] = set()
            for sm in [self, other]:
                if sm._contains(b, page_addr):
                    memory_objects = set()
                    for mo in sm.content_gen(b):
                        if mo.includes(page_addr + b):
                            memory_objects.add(mo)
                    memory_object_sets.add(frozenset(memory_objects))
                    if sm is self:
                        self_has_memory_object_set = True
                else:
                    unconstrained_in.append(sm)

            if not memory_object_sets:
                continue
            if self_has_memory_object_set and len(memory_object_sets) == 1:
                continue

            # TODO: compare_values even more?
            return False

        return True

    def changed_bytes(self, other: MVListPage, page_addr: int | None = None):
        candidates: set[int] = super().changed_bytes(other)
        if candidates is not None:
            # using the result from the history tracking mixin as an approximation
            return candidates

        # slower path
        if candidates is None:
            candidates: set[int] = set()
            # resort to the slower solution
            if self.sinkhole is None:
                candidates |= self.stored_offset
            else:
                for i in range(len(self.content)):
                    if self._contains(i, page_addr):
                        candidates.add(i)

            if other.sinkhole is None:
                candidates |= other.stored_offset
            else:
                for i in range(len(other.content)):
                    if other._contains(i, page_addr):
                        candidates.add(i)

        byte_width = 8  # TODO: Introduce self.state if we want to use self.state.arch.byte_width
        differences: set[int] = set()
        for c in candidates:
            s_contains = self._contains(c, page_addr)
            o_contains = other._contains(c, page_addr)
            if (not s_contains and o_contains) or (s_contains and not o_contains):
                differences.add(c)
            else:
                if self.content[c] is None and self.sinkhole is not None:
                    self.content[c] = SimMemoryObject(
                        self.sinkhole.bytes_at(page_addr + c, 1),
                        page_addr + c,
                        byte_width=byte_width,
                        endness="Iend_BE",
                    )
                if other.content[c] is None and other.sinkhole is not None:
                    other.content[c] = SimMemoryObject(
                        other.sinkhole.bytes_at(page_addr + c, 1),
                        page_addr + c,
                        byte_width=byte_width,
                        endness="Iend_BE",
                    )
                if s_contains and self.content[c] != other.content[c]:
                    same = None
                    if self._mo_cmp is not None:
                        same = self._mo_cmp(self.content[c], other.content[c], page_addr + c, 1)
                    if same is None:
                        # Try to see if the bytes are equal
                        self_bytes = {mo.bytes_at(page_addr + c, 1) for mo in self.content_gen(c)}
                        other_bytes = {mo.bytes_at(page_addr + c, 1) for mo in other.content_gen(c)}
                        same = self_bytes == other_bytes

                    if same is False:
                        differences.add(c)
                else:
                    # this means the byte is in neither memory
                    pass

        return differences

    def content_gen(self, index):
        if self.content[index] is None:
            return
        elif type(self.content[index]) is set:
            yield from self.content[index]
        else:
            yield self.content[index]

    def _contains(self, off: int, page_addr: int):
        if off >= len(self.content):
            return False
        if self.content[off] is not None:
            return True
        if self.sinkhole is None:
            return False
        return self.sinkhole.includes(page_addr + off)

    def _replace_mo(
        self, old_mo: SimMemoryObject, new_mo: SimMemoryObject, page_addr: int, page_size: int
    ) -> SimMemoryObject:
        if self.sinkhole is old_mo:
            self.sinkhole = new_mo
        else:
            start, end = self._resolve_range(old_mo, page_addr, page_size)
            for i in range(start, end):
                s = {new_mo}
                if self.content[i - page_addr] is old_mo:
                    self.content[i - page_addr] = s
        return new_mo

    @staticmethod
    def _resolve_range(mo: SimMemoryObject, page_addr: int, page_size) -> tuple[int, int]:
        start = max(mo.base, page_addr)
        end = min(mo.last_addr + 1, page_addr + page_size)
        if end <= start:
            l.warning("Nothing left of the memory object to store in SimPage.")
        return start, end

    def _get_objects(self, start: int, page_addr: int) -> list[SimMemoryObject] | None:
        mos = self.content[start]
        if mos is None:
            return None
        lst = []
        if type(mos) is set:
            for mo in mos:
                if mo.includes(start + page_addr):
                    lst.append(mo)
        else:
            lst.append(mos)
        if lst:
            return lst
        return None



from __future__ import annotations
import os
import re
import logging
from sortedcontainers import SortedList

from angr.analyses import Analysis, register_analysis
from angr.analyses.decompiler.decompilation_cache import DecompilationCache
from .base import BaseStructuredCodeGenerator, InstructionMapping, PositionMapping
from angr.knowledge_plugins.functions.function import Function


l = logging.getLogger(__name__)


class ImportedLine:
    def __init__(self, addr):
        self.tags = {"ins_addr": addr}


class ImportSourceCode(BaseStructuredCodeGenerator, Analysis):
    def __init__(self, function, flavor="source", source_root=None, encoding="utf-8"):
        super().__init__(flavor=flavor)

        if isinstance(function, (int, str)):
            function = self.kb.functions[function]
        self.function: Function = function
        self._source_root = source_root
        self._encoding = encoding

        self.regenerate_text()

        if flavor is not None and self.text:
            if (function.addr, flavor) not in self.kb.decompilations:
                self.kb.decompilations[(function.addr, flavor)] = DecompilationCache(function.addr)
            self.kb.decompilations[(function.addr, flavor)].codegen = self

    def regenerate_text(self):
        cache = {}
        ranges = self._compute_function_ranges(cache=cache)
        line_to_addr = self._compute_line_to_addr(ranges)

        self.stmt_posmap = PositionMapping()
        self.insmap = InstructionMapping()
        pos = 0
        all_lines = []
        for filename, range_start, range_end in ranges:
            these_lines = self._open_file(filename)[range_start - 1 : range_end - 1 + 1]
            all_lines.extend(these_lines)
            for idx, line_text in enumerate(these_lines):
                addr = line_to_addr[(filename, range_start + idx)]
                if addr is not None:
                    self.stmt_posmap.add_mapping(pos, len(line_text), ImportedLine(addr))
                    self.insmap.add_mapping(addr, pos + len(line_text) - len(line_text.lstrip(" \t")))
                pos += len(line_text)

        self.text = "".join(all_lines)

    def _locate_file(self, filename):
        if os.path.isfile(filename):
            return filename
        if self._source_root is None:
            return None

        filename = filename.strip("/\\")
        path_keys = re.split("[/\\\\]", filename)
        for i in range(len(path_keys)):
            maybe_path = os.path.join(self._source_root, *path_keys[i:])
            if os.path.isfile(maybe_path):
                return maybe_path

        return None

    def _open_file(self, name, cache=None):
        if cache is None:
            cache = {}
        if name in cache:
            return cache[name]
        local_name = self._locate_file(name)
        if local_name is None:
            l.warning("Could not find a local copy of %s", name)
            cache[name] = None
            return None
        with open(local_name, encoding=self._encoding) as fp:
            line_data = fp.readlines()
        cache[name] = line_data
        return line_data

    def _compute_function_ranges(self, cache=None):
        if cache is None:
            cache = {}

        obj = self.project.loader.find_object_containing(self.function.addr)
        if obj is None:
            l.warning("No object contains function %s", self.function.name)
            return []

        if self.function.addr not in obj.addr_to_line:
            l.info("No line data for function %s", self.function.name)
            return []

        lines = SortedList()
        for block in self.function.blocks:
            for insn in block.instruction_addrs:
                if insn in obj.addr_to_line:
                    lines.update((*filename_line, insn) for filename_line in obj.addr_to_line[insn])

        ranges = []
        for filename, line, _ in lines:
            if any(r[0] == filename and r[1] <= line <= r[2] for r in ranges):
                continue
            file_lines = self._open_file(filename, cache=cache)
            if file_lines is None:
                continue
            if line - 1 >= len(file_lines) or line - 1 < 0:
                # a non-existent line number is specified. skip this record
                l.warning(
                    "Line number %d does not exist in file %s. It might be the wrong source code file.", line, filename
                )
                continue
            range_start = line
            range_end = line

            # find the first { on the line
            # find the matching }
            # scan backwards for any lines that aren't blank and don't have {}; on them
            # scan forwards for any lines continued via \
            col = file_lines[range_start - 1].find("{")
            if col != -1:
                col += 1
                stack = 1
                while range_end - 1 < len(file_lines):
                    while col < len(file_lines[range_end - 1]):
                        ch = file_lines[range_end - 1][col]
                        if ch == "{":
                            stack += 1
                        elif ch == "}":
                            stack -= 1
                        if stack == 0:
                            break
                        col += 1
                    else:
                        col = 0
                        range_end += 1
                        continue
                    break

            maybe_prev_line = range_start - 1
            while maybe_prev_line >= 0:
                linedata = file_lines[maybe_prev_line - 1].strip()
                if linedata and not any(c in linedata for c in "{};"):
                    range_start = maybe_prev_line
                    maybe_prev_line -= 1
                else:
                    break

            while range_end - 1 < len(file_lines) - 1:
                linedata = file_lines[range_end - 1].strip()
                if linedata.endswith("\\"):
                    range_end += 1
                else:
                    break

            if any(
                r[0] == filename and (r[1] <= range_start <= r[2] or range_start <= r[1] <= range_end) for r in ranges
            ):
                l.error("Detected line ranges are overlapping?")
            ranges.append((filename, range_start, range_end))

        return ranges

    def _compute_line_to_addr(self, ranges):
        result = {}
        for filename, range_start, range_end in ranges:
            for line in range(range_start, range_end + 1):
                result[(filename, line)] = None

        obj = self.project.loader.find_object_containing(self.function.addr)
        if obj is None:
            l.error("There is a function whose address does not correspond to any loaded object")
            return {}

        for addr, filename_line in obj.addr_to_line.items():
            if filename_line in result:
                result[filename_line] = addr

        return result


register_analysis(ImportSourceCode, "ImportSourceCode")



from __future__ import annotations
from angr.ailment.expression import BinaryOp, UnaryOp, Expression
from angr.ailment.statement import Statement, Assignment
from angr.ailment import Block
from angr.project import Project
from angr.knowledge_base import KnowledgeBase
from angr.knowledge_plugins.key_definitions import atoms


class PeepholeOptimizationStmtBase:
    """
    The base class for all peephole optimizations that are applied on AIL statements.
    """

    __slots__ = (
        "func_addr",
        "kb",
        "preserve_vvar_ids",
        "project",
        "type_hints",
    )
    project: Project | None
    kb: KnowledgeBase | None
    func_addr: int | None
    preserve_vvar_ids: set[int]
    type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]]

    NAME = "Peephole Optimization - Statement"
    DESCRIPTION = "Peephole Optimization - Statement"
    stmt_classes = None

    def __init__(
        self,
        project: Project | None,
        kb: KnowledgeBase | None,
        func_addr: int | None = None,
        preserve_vvar_ids: set[int] | None = None,
        type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]] | None = None,
    ):
        self.project = project
        self.kb = kb
        self.func_addr = func_addr
        self.preserve_vvar_ids = set() if preserve_vvar_ids is None else preserve_vvar_ids
        self.type_hints = [] if type_hints is None else type_hints

    def optimize(self, stmt, stmt_idx: int | None = None, block=None, **kwargs):
        raise NotImplementedError("_optimize() is not implemented.")


class PeepholeOptimizationMultiStmtBase:
    """
    The base class for all peephole optimizations that are applied on multiple AIL statements at once.
    """

    __slots__ = (
        "func_addr",
        "kb",
        "preserve_vvar_ids",
        "project",
        "type_hints",
    )
    project: Project | None
    kb: KnowledgeBase | None
    func_addr: int | None
    preserve_vvar_ids: set[int]
    type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]]

    NAME = "Peephole Optimization - Multi-statement"
    DESCRIPTION = "Peephole Optimization - Multi-statement"
    stmt_classes = None

    def __init__(
        self,
        project: Project | None,
        kb: KnowledgeBase | None,
        func_addr: int | None = None,
        preserve_vvar_ids: set[int] | None = None,
        type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]] | None = None,
    ):
        self.project = project
        self.kb = kb
        self.func_addr = func_addr
        self.preserve_vvar_ids = set() if preserve_vvar_ids is None else preserve_vvar_ids
        self.type_hints = [] if type_hints is None else type_hints

    def optimize(self, stmts: list[Statement], stmt_idx: int | None = None, block=None, **kwargs):
        raise NotImplementedError("_optimize() is not implemented.")


class PeepholeOptimizationExprBase:
    """
    The base class for all peephole optimizations that are applied on AIL expressions.
    """

    __slots__ = (
        "func_addr",
        "kb",
        "preserve_vvar_ids",
        "project",
        "type_hints",
    )
    project: Project | None
    kb: KnowledgeBase | None
    func_addr: int | None
    preserve_vvar_ids: set[int]
    type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]]

    NAME = "Peephole Optimization - Expression"
    DESCRIPTION = "Peephole Optimization - Expression"
    expr_classes = None

    def __init__(
        self,
        project: Project | None,
        kb: KnowledgeBase | None,
        func_addr: int | None = None,
        preserve_vvar_ids: set[int] | None = None,
        type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]] | None = None,
    ):
        self.project = project
        self.kb = kb
        self.func_addr = func_addr
        self.preserve_vvar_ids = set() if preserve_vvar_ids is None else preserve_vvar_ids
        self.type_hints = [] if type_hints is None else type_hints

    def optimize(self, expr, stmt_idx: int | None = None, block=None, **kwargs):
        raise NotImplementedError("_optimize() is not implemented.")

    #
    # Util methods
    #

    @staticmethod
    def find_definition(ail_expr: Expression, stmt_idx: int, block: Block) -> None:
        idx = stmt_idx - 1
        if idx >= 0:
            stmt = block.statements[idx]
            if isinstance(stmt, Assignment) and stmt.dst.likes(ail_expr):
                return stmt.src
        return None

    @staticmethod
    def is_bool_expr(ail_expr):
        if isinstance(ail_expr, BinaryOp) and ail_expr.op in {
            "CmpEQ",
            "CmpNE",
            "CmpLT",
            "CmpLE",
            "CmpGT",
            "CmpGE",
            "CmpLTs",
            "CmpLEs",
            "CmpGTs",
            "CmpGEs",
        }:
            return True
        return bool(isinstance(ail_expr, UnaryOp) and ail_expr.op == "Not")



from __future__ import annotations

import claripy

from angr.errors import SimMemoryError
from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class SimpleInterfaceMixin(MemoryMixin):
    def load(self, addr, size=None, *, endness=None, condition=None, fallback=None, **kwargs):
        tsize = self._translate_size(size, None)
        return super().load(
            self._translate_addr(addr),
            size=tsize,
            endness=self._translate_endness(endness),
            condition=self._translate_cond(condition),
            fallback=self._translate_data(fallback, tsize) if fallback is not None else None,
            **kwargs,
        )

    def store(self, addr, data, size=None, *, endness=None, condition=None, **kwargs):
        tsize = self._translate_size(size, data)
        super().store(
            self._translate_addr(addr),
            self._translate_data(data, tsize),
            size=tsize,
            endness=self._translate_endness(endness),
            condition=self._translate_cond(condition),
            **kwargs,
        )

    def _translate_addr(self, a):
        if isinstance(a, int):
            return a
        if isinstance(a, claripy.ast.Base) and not a.singlevalued:
            raise SimMemoryError("address not supported")
        return self.state.solver.eval(a)

    def _translate_data(self, d, size):
        if type(d) in (bytes, bytearray):
            return claripy.BVV(d)
        if type(d) is int:
            return claripy.BVV(d, size * self.state.arch.byte_width)
        if isinstance(d, claripy.ast.Base):
            return d
        raise SimMemoryError("data not supported")

    def _translate_size(self, s, data):
        if isinstance(s, int):
            return s
        if isinstance(s, claripy.ast.Base) and not s.singlevalued:
            raise SimMemoryError("size not supported")
        if s is None:
            if isinstance(data, claripy.ast.BV):
                return len(data) // self.state.arch.byte_width
            if isinstance(data, (bytes, bytearray)):
                return len(data)
            raise SimMemoryError("unknown size")
        return self.state.solver.eval(s)

    def _translate_cond(self, c):
        if isinstance(c, claripy.ast.Base) and not c.singlevalued:
            raise SimMemoryError("condition not supported")
        if c is None:
            return True
        return self.state.solver.eval_upto(c, 1)[0]

    def _translate_endness(self, endness):
        if endness is None:
            return self.endness
        return endness



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
import math

from angr.ailment.expression import Convert, BinaryOp, Const, Expression

from .base import PeepholeOptimizationExprBase


class OptimizedDivisionSimplifier(PeepholeOptimizationExprBase):
    """
    Convert expressions with right shifts into expressions with divisions.
    """

    __slots__ = ()

    NAME = "Simplify optimized division expressions, e.g., (N * a) >> M => a / N1"
    expr_classes = (Convert, BinaryOp)

    def optimize(  # pylint:disable=unused-argument
        self, expr: Convert | BinaryOp, stmt_idx: int | None = None, block=None, **kwargs
    ):
        r = None

        if isinstance(expr, Convert):
            if expr.from_bits == 64 and expr.to_bits == 32 and isinstance(expr.operand, BinaryOp):
                r = self.optimize_binaryop(expr.operand)

        elif isinstance(expr, BinaryOp):
            r = self.optimize_binaryop(expr)

        # keep size
        if r is not None and r.bits < expr.bits:
            r = Convert(expr.idx, r.bits, expr.bits, False, r, **expr.tags)

        return r

    def optimize_binaryop(self, original_expr: BinaryOp):
        if isinstance(original_expr, BinaryOp):
            # try to unify if both operands are wrapped with Convert()
            conv_expr = self._unify_conversion(original_expr)
            expr = original_expr if conv_expr is None else conv_expr.operand
            assert isinstance(expr, BinaryOp)

            if expr.op == "Shr" and isinstance(expr.operands[1], Const):
                r = self._match_case_b(expr)
                if r is not None:
                    return self._reconvert(r, conv_expr) if conv_expr is not None else r
                assert isinstance(expr.operands[1].value, int)
                r = self._match_case_c(expr.operands[0], expr.operands[1].value)
                if r is not None:
                    return self._reconvert(r, conv_expr) if conv_expr is not None else r

            elif expr.op in {"Add", "Sub"}:
                expr0, expr1 = expr.operands
                if isinstance(expr1, Convert) and expr1.from_bits == 32 and expr1.to_bits == 64:
                    r = self._match_case_a(expr0, expr1)
                    if r is not None:
                        return self._reconvert(r, conv_expr) if conv_expr is not None else r

                # with Convert in consideration
                if (
                    isinstance(expr0, BinaryOp)
                    and expr0.op in {"Shr", "Sar"}
                    and isinstance(expr0.operands[1], Const)
                    and expr0.operands[1].value == 0x3F
                    and isinstance(expr1, BinaryOp)
                    and expr1.op in {"Shr", "Sar"}
                    and isinstance(expr1.operands[1], Const)
                    and expr1.operands[1].value == 0x20
                ):
                    expr0_operand = expr0.operands[0]
                    expr1_operand = expr1.operands[0]
                    if (
                        isinstance(expr0_operand, BinaryOp)
                        and expr0_operand.op in {"Mull", "Mul"}
                        and expr0_operand.signed
                        and isinstance(expr0_operand.operands[1], Const)
                    ):
                        a0 = expr0_operand.operands[0]
                        a1 = expr1_operand.operands[0]
                    elif (
                        isinstance(expr1_operand, BinaryOp)
                        and expr1_operand.op in {"Mull", "Mul"}
                        and expr1_operand.signed
                        and isinstance(expr1_operand.operands[1], Const)
                    ):
                        a0 = expr1_operand.operands[0]
                        a1 = expr0_operand.operands[0]
                    else:
                        a0, a1 = None, None

                    if a0 is not None and a1 is not None and a0.likes(a1):
                        # (a * x >> 0x3f) +/- (a * x >> 0x20)  ==>  a / N
                        C = expr0_operand.operands[1].value
                        X = a0
                        V = 32
                        ndigits = 5 if V == 32 else 6
                        divisor = self._check_divisor(pow(2, V), C, ndigits)
                        if divisor is not None:
                            new_const = Const(None, None, divisor, X.bits)
                            r = BinaryOp(
                                expr0_operand.idx,
                                "Div",
                                [X, new_const],
                                expr0_operand.signed,
                                **expr0_operand.tags,
                            )
                            return self._reconvert(r, conv_expr) if conv_expr is not None else r

        return None

    def _match_case_a(self, expr0: Expression, expr1: Convert) -> BinaryOp | None:
        # (
        #   (((Conv(32->64, vvar_44{reg 32}) * 0x4325c53f<64>) >>a 0x24<8>) & 0xffffffff<64>) -
        #   Conv(32->s64, (vvar_44{reg 32} >>a 0x1f<8>))
        # )

        expr1_op = expr1.operand

        if (
            isinstance(expr0, BinaryOp)
            and expr0.op == "And"
            and isinstance(expr0.operands[1], Const)
            and expr0.operands[1].value == 0xFFFFFFFF
        ):
            expr0 = expr0.operands[0]
        else:
            return None

        if (
            isinstance(expr0, BinaryOp)
            and expr0.op in {"Shr", "Sar"}
            and isinstance(expr0.operands[1], Const)
            and isinstance(expr1_op, BinaryOp)
            and expr1_op.op in {"Shr", "Sar"}
            and isinstance(expr1_op.operands[1], Const)
        ):
            if (
                isinstance(expr0.operands[0], BinaryOp)
                and expr0.operands[0].op in {"Mull", "Mul"}
                and isinstance(expr0.operands[0].operands[1], Const)
            ):
                a0 = expr0.operands[0].operands[0]
                a1 = expr1_op.operands[0]
            elif (
                isinstance(expr1_op.operands[0], BinaryOp)
                and expr1_op.operands[0].op in {"Mull", "Mul"}
                and isinstance(expr1_op.operands[0].operands[1], Const)
            ):
                a1 = expr0.operands[0].operands[0]
                a0 = expr1_op.operands[0]
            else:
                a0, a1 = None, None

            # a0: Conv(32->64, vvar_44{reg 32})
            # a1: vvar_44{reg 32}
            if isinstance(a0, Convert) and a1 is not None and a0.from_bits == a1.bits:
                a0 = a0.operand

            if a0 is not None and a1 is not None and a0.likes(a1):
                # (a * x >> M1) +/- (a >> M2)  ==>  a / N
                C = expr0.operands[0].operands[1].value
                X = a0
                V = expr0.operands[1].value
                ndigits = 5 if V == 32 else 6
                divisor = self._check_divisor(pow(2, V), C, ndigits)
                if divisor is not None:
                    new_const = Const(None, None, divisor, X.bits)
                    # we cannot drop the convert in this case
                    return BinaryOp(
                        expr0.operands[0].idx,
                        "Div",
                        [X, new_const],
                        expr0.operands[0].signed,
                        **expr0.operands[0].tags,
                    )

        return None

    @staticmethod
    def _match_case_b(expr: BinaryOp) -> BinaryOp | Convert | None:
        """
        A more complex (but general) case for unsigned 32-bit division by a constant integer.

        Ref: https://ridiculousfish.com/blog/posts/labor-of-division-episode-i.html

        Given n and d, n//d (unsigned) can be rewritten to t >> (p - 1) where
        - p = ceiling(log2(d))
        - m = ceiling((2 ** (32 + p) / d))
        - q = (m * n) >> 32
        - t = q + ((n - q) >> 1)

        We can match the expression against  t >> (p - 1).
        """

        # t >> (p - 1)
        if not (isinstance(expr, BinaryOp) and expr.op == "Shr"):
            return None
        if not (isinstance(expr.operands[1], Const) and expr.operands[1].value > 0):
            return None
        p_minus_1 = expr.operands[1].value
        p = p_minus_1 + 1
        t = expr.operands[0]

        # unmask
        if isinstance(t, BinaryOp) and t.op == "And":
            if isinstance(t.operands[1], Const) and t.operands[1].value == 0xFFFFFFFF:
                t = t.operands[0]
            elif isinstance(t.operands[0], Const) and t.operands[0].value == 0xFFFFFFFF:
                t = t.operands[1]
            else:
                return None

        # t = q + ((n - q) >> 1)
        if not (isinstance(t, BinaryOp) and t.op == "Add"):
            return None

        if (
            isinstance(t.operands[0], BinaryOp)
            and t.operands[0].op == "Shr"
            and isinstance(t.operands[0].operands[1], Const)
            and t.operands[0].operands[1].value == 1
        ):
            q = t.operands[1]
            n_minus_q = t.operands[0].operands[0]
        elif (
            isinstance(t.operands[1], BinaryOp)
            and t.operands[1].op == "Shr"
            and isinstance(t.operands[1].operands[1], Const)
            and t.operands[1].operands[1].value == 1
        ):
            q = t.operands[0]
            n_minus_q = t.operands[1]
        else:
            return None
        if isinstance(q, Convert) and q.from_bits == 64 and q.to_bits == 32:
            q = q.operand
        if isinstance(n_minus_q, Convert) and n_minus_q.from_bits == 64 and n_minus_q.to_bits == 32:
            n_minus_q = n_minus_q.operand

        # unmask
        if isinstance(n_minus_q, BinaryOp) and n_minus_q.op == "And":
            if isinstance(n_minus_q.operands[1], Const) and n_minus_q.operands[1].value == 0xFFFFFFFF:
                n_minus_q = n_minus_q.operands[0]
            elif isinstance(n_minus_q.operands[0], Const) and n_minus_q.operands[0].value == 0xFFFFFFFF:
                n_minus_q = n_minus_q.operands[1]
            else:
                return None

        if not (isinstance(n_minus_q, BinaryOp) and n_minus_q.op == "Sub"):
            return None
        if not q.likes(n_minus_q.operands[1]):
            return None

        # q = (m * n) >> 32
        if not (
            isinstance(q, BinaryOp) and q.op == "Shr" and isinstance(q.operands[1], Const) and q.operands[1].value == 32
        ):
            return None
        if not (isinstance(q.operands[0], BinaryOp) and q.operands[0].op in {"Mull", "Mul"}):
            return None
        if isinstance(q.operands[0].operands[1], Const):
            n = q.operands[0].operands[0]
            m = q.operands[0].operands[1].value
        elif isinstance(q.operands[0].operands[0], Const):
            n = q.operands[0].operands[1]
            m = q.operands[0].operands[0].value
        else:
            # this should never happen, because multiplication of two constants are eagerly evaluated
            return None

        assert isinstance(m, int) and isinstance(p, int)
        divisor = math.ceil((2 ** (32 + p)) / (m + 0x1_0000_0000))
        if divisor == 0:
            return None
        divisor_expr = Const(None, None, divisor, n.bits)
        div = BinaryOp(expr.idx, "Div", [n, divisor_expr], signed=False, **expr.tags)
        if expr.bits != div.bits:
            div = Convert(expr.idx, div.bits, expr.bits, False, div, **expr.tags)
        return div

    def _match_case_c(self, inner, m: int) -> BinaryOp | None:
        # (N * a) >> M  ==>  a / N1
        if isinstance(inner, BinaryOp) and inner.op in {"Mull", "Mul"} and not inner.signed:
            if isinstance(inner.operands[0], Const) and not isinstance(inner.operands[1], Const):
                C = inner.operands[0].value
                X = inner.operands[1]
            elif isinstance(inner.operands[1], Const) and not isinstance(inner.operands[0], Const):
                C = inner.operands[1].value
                X = inner.operands[0]
            else:
                C = X = None

            if C is not None and X is not None:
                V = m
                ndigits = 5 if V == 32 else 6
                divisor = self._check_divisor(pow(2, V), C, ndigits)
                if divisor is not None:
                    new_const = Const(None, None, divisor, X.bits)
                    return BinaryOp(inner.idx, "Div", [X, new_const], inner.signed, **inner.tags)
        return None

    @staticmethod
    def _check_divisor(a, b, ndigits=6):
        if b == 0:
            return None
        divisor_1 = 1 + (a // b)
        divisor_2 = int(round(a / float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None

    @staticmethod
    def _unify_conversion(expr: BinaryOp) -> Convert | None:
        if (
            isinstance(expr.operands[0], Convert)
            and isinstance(expr.operands[1], Convert)
            and expr.operands[1].to_bits == expr.operands[0].to_bits
            and expr.operands[1].from_bits == expr.operands[0].from_bits
            and expr.op in {"Add", "Sub"}
        ):
            op0 = expr.operands[0]
            op0_inner = expr.operands[0].operand
            # op1 = expr.operands[1]
            op1_inner = expr.operands[1].operand

            new_expr = BinaryOp(
                expr.idx,
                expr.op,
                (op0_inner, op1_inner),
                expr.signed,
                bits=op0.from_bits,
                **expr.tags,
            )
            return Convert(
                op0.idx,
                op0.from_bits,
                op0.to_bits,
                op0.is_signed,
                new_expr,
                **op0.tags,
            )
        return None

    @staticmethod
    def _reconvert(expr: Expression, conv: Convert) -> Convert:
        return Convert(
            conv.idx,
            conv.from_bits,
            conv.to_bits,
            conv.is_signed,
            expr,
            from_type=conv.from_type,
            to_type=conv.to_type,
            rounding_mode=conv.rounding_mode,
            **conv.tags,
        )



# pylint:disable=missing-class-docstring,too-many-boolean-expressions
from __future__ import annotations
from typing import Any, TYPE_CHECKING
import string
import logging

import capstone
import networkx

import claripy

from angr.analyses import Analysis, AnalysesHub
from angr.errors import SimMemoryMissingError, AngrCallableMultistateError, AngrCallableError, AngrAnalysisError
from angr.calling_conventions import SimRegArg, default_cc
from angr.state_plugins.sim_action import SimActionData
from angr.sim_options import ZERO_FILL_UNCONSTRAINED_REGISTERS, ZERO_FILL_UNCONSTRAINED_MEMORY, TRACK_MEMORY_ACTIONS
from angr.sim_type import SimTypeFunction, SimTypeBottom, SimTypePointer
from angr.analyses.reaching_definitions import ObservationPointType
from angr.utils.graph import GraphUtils

from .irsb_reg_collector import IRSBRegisterCollector

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function


_l = logging.getLogger(__name__)


STEP_LIMIT_FIND = 500
STEP_LIMIT_ANALYSIS = 5000


ALL_X64_XMM_REGS = {
    capstone.x86.X86_REG_XMM0,
    capstone.x86.X86_REG_XMM1,
    capstone.x86.X86_REG_XMM2,
    capstone.x86.X86_REG_XMM3,
    capstone.x86.X86_REG_XMM4,
    capstone.x86.X86_REG_XMM5,
    capstone.x86.X86_REG_XMM6,
    capstone.x86.X86_REG_XMM7,
    capstone.x86.X86_REG_XMM8,
    capstone.x86.X86_REG_XMM9,
    capstone.x86.X86_REG_XMM10,
    capstone.x86.X86_REG_XMM11,
    capstone.x86.X86_REG_XMM12,
    capstone.x86.X86_REG_XMM13,
    capstone.x86.X86_REG_XMM14,
    capstone.x86.X86_REG_XMM15,
    capstone.x86.X86_REG_XMM16,
    capstone.x86.X86_REG_XMM17,
    capstone.x86.X86_REG_XMM18,
    capstone.x86.X86_REG_XMM19,
    capstone.x86.X86_REG_XMM20,
    capstone.x86.X86_REG_XMM21,
    capstone.x86.X86_REG_XMM22,
    capstone.x86.X86_REG_XMM23,
    capstone.x86.X86_REG_XMM24,
    capstone.x86.X86_REG_XMM25,
    capstone.x86.X86_REG_XMM26,
    capstone.x86.X86_REG_XMM27,
    capstone.x86.X86_REG_XMM28,
    capstone.x86.X86_REG_XMM29,
    capstone.x86.X86_REG_XMM30,
    capstone.x86.X86_REG_XMM31,
}


class StringDeobFuncDescriptor:
    """
    Describes a string deobfuscation function.
    """

    string_input_arg_idx: int
    string_output_arg_idx: int
    string_length_arg_idx: int | None
    string_null_terminating: bool | None

    def __init__(self):
        self.string_length_arg_idx = None
        self.string_null_terminating = None


class StringObfuscationFinder(Analysis):
    """
    An analysis that automatically finds string obfuscation routines.
    """

    def __init__(self, functions: list[Function] | None = None):
        self._functions_to_analyze = functions
        self._function_addrs_to_analyze = None if functions is None else {f.addr for f in functions}

        self.type1_candidates = []
        self.type2_candidates = []
        self.type3_candidates = []

        self.analyze()

    def analyze(self):
        _l.debug("Finding type 1 candidates.")
        self.type1_candidates = self._find_type1()
        _l.debug("Got %d type 1 candidates.", len(self.type1_candidates))

        _l.debug("Finding type 2 candidates.")
        self.type2_candidates = self._find_type2()
        _l.debug("Got %d type 2 candidates.", len(self.type2_candidates))

        _l.debug("Finding type 3 candidates.")
        self.type3_candidates = self._find_type3()
        _l.debug("Got %d type 3 candidates.", len(self.type3_candidates))
        _l.debug("Done.")

        if self.type1_candidates:
            for type1_func_addr, desc in self.type1_candidates:
                _l.debug("Analyzing type 1 candidates.")
                type1_deobfuscated, type1_string_loader_candidates = self._analyze_type1(type1_func_addr, desc)
                self.kb.obfuscations.type1_deobfuscated_strings.update(type1_deobfuscated)
                self.kb.obfuscations.type1_string_loader_candidates |= type1_string_loader_candidates

        if self.type2_candidates:
            for type2_func_addr, desc, string_candidates in self.type2_candidates:
                _l.debug("Analyzing type 2 candidates.")
                type2_string_loader_candidates = self._analyze_type2(
                    type2_func_addr, desc, {addr for addr, _, _ in string_candidates}
                )
                type2_deobfuscated_strings = {addr: s for addr, _, s in string_candidates}
                self.kb.obfuscations.type2_deobfuscated_strings.update(type2_deobfuscated_strings)
                self.kb.obfuscations.type2_string_loader_candidates |= type2_string_loader_candidates

        if self.type3_candidates:
            for type3_func_addr, desc in self.type3_candidates:
                _l.debug("Analyzing type 3 candidates.")
                type3_strings = self._analyze_type3(type3_func_addr, desc)
                self.kb.obfuscations.type3_deobfuscated_strings.update(type3_strings)

    def _find_type1(self) -> list[tuple[int, StringDeobFuncDescriptor]]:
        # Type 1 string deobfuscation functions
        # - Take a constant string or local string as input
        # - Output strings that are reasonable
        # - Do not call other functions (i.e., these functions are leaf functions)
        #
        # Type 1 string deobfuscation functions will decrypt each string once and for good.

        cfg = self.kb.cfgs.get_most_accurate()
        if cfg is None:
            raise AngrAnalysisError("StringObfuscationFinder needs a CFG for the analysis")

        arch = self.project.arch

        type1_candidates: list[tuple[int, StringDeobFuncDescriptor]] = []

        functions_to_analyze = (
            self.project.kb.functions.values() if self._functions_to_analyze is None else self._functions_to_analyze
        )
        for func in functions_to_analyze:
            if func.is_simprocedure or func.is_plt or func.is_alignment:
                continue

            if func.prototype is None or len(func.prototype.args) < 1:
                continue

            if len(func.arguments) != len(func.prototype.args):
                # function argument locations and function prototype arguments do not match
                continue

            if self.project.kb.functions.callgraph.out_degree[func.addr] != 0:
                continue

            # find its callsites and arguments
            callers = [
                pred for pred in self.project.kb.functions.callgraph.predecessors(func.addr) if pred != func.addr
            ]

            if not callers:
                continue

            if len(func.block_addrs_set) <= 2:
                # function is too small...
                continue
            if len(func.block_addrs_set) >= 50:
                # function is too big...
                continue

            # decompile this function and see if it "looks like" a deobfuscation function
            with self._resilience():
                dec = self.project.analyses.Decompiler(func, cfg=cfg, fail_fast=self._fail_fast)  # type:ignore
            if (
                dec.codegen is None
                or not dec.codegen.text
                or not self._like_type1_deobfuscation_function(dec.codegen.text)
            ):
                continue

            func_node = cfg.get_any_node(func.addr)
            if func_node is None:
                continue

            args_list = []
            for caller in callers:
                callsite_nodes = [
                    pred
                    for pred in cfg.get_predecessors(func_node)
                    if pred.function_address == caller and pred.instruction_addrs
                ]
                observation_points = []
                for callsite_node in callsite_nodes:
                    observation_points.append(
                        ("insn", callsite_node.instruction_addrs[-1], ObservationPointType.OP_BEFORE)
                    )
                rda = self.project.analyses.ReachingDefinitions(
                    self.project.kb.functions[caller],
                    observe_all=False,
                    observation_points=observation_points,
                )
                for callsite_node in callsite_nodes:
                    observ = rda.model.get_observation_by_insn(
                        callsite_node.instruction_addrs[-1],
                        ObservationPointType.OP_BEFORE,
                    )
                    if observ is None:
                        continue
                    # load values for each function argument
                    args: list[tuple[int, Any]] = []
                    for arg_idx, func_arg in enumerate(func.arguments):
                        # FIXME: We are ignoring all non-register function arguments until we see a test case where
                        # FIXME: stack-passing arguments are used
                        real_arg = func.prototype.args[arg_idx]
                        if isinstance(func_arg, SimRegArg):
                            reg_offset, reg_size = arch.registers[func_arg.reg_name]
                            arg_size = (
                                real_arg.size if real_arg.size is not None else reg_size
                            ) // self.project.arch.byte_width
                            try:
                                mv = observ.registers.load(reg_offset, size=arg_size)
                            except SimMemoryMissingError:
                                args.append((arg_idx, claripy.BVV(0xDEADBEEF, self.project.arch.bits)))
                                continue
                            arg_value = mv.one_value()
                            if arg_value is None:
                                arg_value = claripy.BVV(0xDEADBEEF, self.project.arch.bits)
                            args.append((arg_idx, arg_value))

                    # the args must have at least one concrete address that points to an initialized memory location
                    acceptable_args = False
                    for _, arg in args:
                        if arg is not None and arg.concrete:
                            v = arg.concrete_value
                            section = self.project.loader.find_section_containing(v)
                            if section is not None:
                                acceptable_args = True
                                break
                    if acceptable_args:
                        args_list.append(args)

            if not args_list:
                continue

            is_candidate = False
            desc = StringDeobFuncDescriptor()
            # now that we have good arguments, let's test the function!
            for args in args_list:
                func_call = self.project.factory.callable(
                    func.addr,
                    concrete_only=True,
                    cc=func.calling_convention,
                    prototype=func.prototype,
                    add_options={
                        ZERO_FILL_UNCONSTRAINED_MEMORY,
                        ZERO_FILL_UNCONSTRAINED_REGISTERS,
                    },
                    step_limit=STEP_LIMIT_FIND,
                )

                # before calling the function, let's record the crime scene
                values: list[tuple[int, int, bytes]] = []
                for arg_idx, arg in args:
                    if arg is not None and arg.concrete:
                        v = arg.concrete_value
                        section = self.project.loader.find_section_containing(v)
                        if section is not None:
                            values.append((arg_idx, v, self.project.loader.memory.load(v, 100)))

                try:
                    func_call(*[arg for _, arg in args])
                except (AngrCallableMultistateError, AngrCallableError):
                    continue

                if func_call.result_state is None:
                    continue

                # let's see what this amazing function has done
                # TODO: Support cases where input and output are using different function arguments
                for arg_idx, addr, old_value in values:
                    out = func_call.result_state.solver.eval(
                        func_call.result_state.memory.load(addr, size=len(old_value)), cast_to=bytes
                    )
                    if out == old_value:
                        continue
                    if self._is_string_reasonable(out):
                        # found it!
                        _l.debug("[+] Deobfuscated string by function %s: %s", repr(func), out)
                        is_candidate = True
                        desc.string_input_arg_idx = arg_idx
                        desc.string_output_arg_idx = arg_idx
                        desc.string_null_terminating = True  # FIXME
                        break

            if is_candidate:
                type1_candidates.append((func.addr, desc))

        return type1_candidates

    def _analyze_type1(self, func_addr: int, desc: StringDeobFuncDescriptor) -> tuple[dict, set]:
        """
        Analyze Type 1 string deobfuscation functions, determine the following information:

        - Deobfuscated strings, lengths, and their addresses
        - Functions that load deobfuscated strings

        :param func_addr:
        :param desc:
        :return:
        """

        deobfuscated_strings = {}

        arch = self.project.arch
        cfg = self.kb.cfgs.get_most_accurate()
        if cfg is None:
            raise AngrAnalysisError("StringObfuscationFinder needs a CFG for the analysis")

        func = self.kb.functions.get_by_addr(func_addr)
        func_node = cfg.get_any_node(func_addr)
        assert func_node is not None
        # Find all call sites for this function
        call_sites = cfg.get_predecessors(func_node)
        rda_cache = {}
        for callsite_node in call_sites:
            # dump arguments
            if callsite_node.function_address in rda_cache:
                rda = rda_cache[callsite_node.function_address]
            else:
                rda = self.project.analyses.ReachingDefinitions(
                    self.project.kb.functions[callsite_node.function_address],
                    observe_all=True,
                ).model
                rda_cache[callsite_node.function_address] = rda
            observ = rda.get_observation_by_insn(
                callsite_node.instruction_addrs[-1],
                ObservationPointType.OP_BEFORE,
            )
            if observ is None:
                continue
            args = []
            assert func.prototype is not None and len(func.arguments) == len(func.prototype.args)
            for func_arg, real_arg in zip(func.arguments, func.prototype.args):
                # FIXME: We are ignoring all non-register function arguments until we see a test case where
                # FIXME: stack-passing arguments are used
                if isinstance(func_arg, SimRegArg):
                    reg_offset, reg_size = arch.registers[func_arg.reg_name]
                    arg_size = (
                        real_arg.size if real_arg.size is not None else reg_size
                    ) // self.project.arch.byte_width
                    try:
                        mv = observ.registers.load(reg_offset, size=arg_size)
                    except SimMemoryMissingError:
                        args.append(claripy.BVV(0xDEADBEEF, self.project.arch.bits))
                        continue
                    v = mv.one_value()
                    if v is not None and v.concrete:
                        args.append(v)
                    else:
                        args.append(claripy.BVV(0xDEADBEEF, self.project.arch.bits))

            if None in args:
                _l.debug(
                    "At least one argument cannot be concretized. Skip the call at %#x.",
                    callsite_node.instruction_addrs[-1],
                )
                continue

            # call the function
            func_call = self.project.factory.callable(
                func.addr,
                concrete_only=True,
                cc=func.calling_convention,
                prototype=func.prototype,
                add_options={ZERO_FILL_UNCONSTRAINED_MEMORY, ZERO_FILL_UNCONSTRAINED_REGISTERS},
                step_limit=STEP_LIMIT_ANALYSIS,
            )
            try:
                func_call(*args)
            except AngrCallableMultistateError:
                _l.debug(
                    "State branching encountered during string deobfuscation. Skip the call at %#x.",
                    callsite_node.instruction_addrs[-1],
                )
                continue
            except AngrCallableError:
                _l.debug(
                    "No path returned. Skip the call at %#x.",
                    callsite_node.instruction_addrs[-1],
                )
                continue

            if func_call.result_state is None:
                continue

            # dump the decrypted string!
            output_addr = args[desc.string_output_arg_idx]
            length = args[desc.string_length_arg_idx].concrete_value if desc.string_length_arg_idx is not None else 256
            output_str = func_call.result_state.solver.eval(
                func_call.result_state.memory.load(output_addr, size=length),
                cast_to=bytes,
            )
            if desc.string_null_terminating and b"\x00" in output_str:
                output_str = output_str[: output_str.index(b"\x00")]
            deobfuscated_strings[output_addr.concrete_value] = output_str

        # for each deobfuscated string, we find its string loader function
        # an obvious candidate function is 0x140001ae4
        xrefs = self.kb.xrefs
        string_loader_candidates = set()
        for str_addr in deobfuscated_strings:
            xref_set = xrefs.get_xrefs_by_dst(str_addr)
            block_addrs = {xref.block_addr for xref in xref_set}
            for block_addr in block_addrs:
                if block_addr is None:
                    continue
                node = cfg.get_any_node(block_addr)
                if node is not None:
                    callees = list(self.kb.functions.callgraph.successors(node.function_address))
                    if callees:
                        #  string loader function should not call anything else
                        continue
                    string_loader_candidates.add(node.function_address)

        return deobfuscated_strings, string_loader_candidates

    def _find_type2(self) -> list[tuple[int, StringDeobFuncDescriptor, list[tuple[int, int, bytes]]]]:
        # Type 2 string deobfuscation functions
        # - Deobfuscates an entire table of encrypted strings
        # - May or may not take any arguments. All arguments should be concrete.
        #
        # Type 2 string deobfuscation functions will decrypt each string once and for good.

        cfg = self.kb.cfgs.get_most_accurate()
        if cfg is None:
            raise AngrAnalysisError("StringObfuscationFinder needs a CFG for the analysis")

        type2_candidates: list[tuple[int, StringDeobFuncDescriptor, list[tuple[int, int, bytes]]]] = []

        functions_to_analyze = (
            self.project.kb.functions.values() if self._functions_to_analyze is None else self._functions_to_analyze
        )
        for func in functions_to_analyze:
            if func.is_simprocedure or func.is_plt or func.is_alignment:
                continue

            if func.prototype is None or len(func.prototype.args) > 1:
                # FIXME: Handle deobfuscation functions that take arguments. Find such a case first
                continue

            if self.project.kb.functions.callgraph.out_degree[func.addr] != 0:
                continue

            # find its callsites and arguments
            callers = [
                pred for pred in self.project.kb.functions.callgraph.predecessors(func.addr) if pred != func.addr
            ]

            if not callers:
                continue

            if len(func.block_addrs_set) <= 2:
                # function is too small...
                continue
            if len(func.block_addrs_set) >= 50:
                # function is too big...
                continue

            # decompile this function and see if it "looks like" a deobfuscation function
            with self._resilience():
                dec = self.project.analyses.Decompiler(
                    func,
                    cfg=cfg,
                    expr_collapse_depth=64,
                    fail_fast=self._fail_fast,  # type:ignore
                )
            if (
                dec.codegen is None
                or not dec.codegen.text
                or not self._like_type2_deobfuscation_function(dec.codegen.text)
            ):
                continue

            desc = StringDeobFuncDescriptor()
            # now that we have good arguments, let's test the function!
            func_call = self.project.factory.callable(
                func.addr,
                concrete_only=True,
                cc=func.calling_convention,
                prototype=func.prototype,
                add_options={TRACK_MEMORY_ACTIONS, ZERO_FILL_UNCONSTRAINED_MEMORY, ZERO_FILL_UNCONSTRAINED_REGISTERS},
                step_limit=STEP_LIMIT_FIND,
            )

            try:
                func_call()
            except (AngrCallableMultistateError, AngrCallableError):
                continue

            if func_call.result_state is None:
                continue

            # where are the reads and writes?
            all_global_reads = []
            all_global_writes = []
            for action in func_call.result_state.history.actions:
                if not isinstance(action, SimActionData):
                    continue
                if not action.actual_addrs:
                    if action.addr is None or not action.addr.ast.concrete:
                        continue
                    actual_addrs = [action.addr.ast.concrete_value]
                else:
                    actual_addrs = action.actual_addrs
                if action.type == "mem":
                    if action.action == "read":
                        assert action.size is not None
                        for a in actual_addrs:
                            for size in range(action.size.ast // 8):
                                all_global_reads.append(a + size)
                    elif action.action == "write":
                        assert action.size is not None
                        for a in actual_addrs:
                            for size in range(action.size.ast // 8):
                                all_global_writes.append(a + size)

            # find likely memory access regions
            all_global_reads = sorted(set(all_global_reads))
            all_global_writes = sorted(set(all_global_writes))
            all_global_write_set = set(all_global_writes)
            # TODO: Handle cases where reads and writes are not going to the same place
            region_candidates: list[tuple[int, int, bytes]] = []
            idx = 0
            while idx < len(all_global_reads):
                starting_offset = all_global_reads[idx]
                if starting_offset not in all_global_write_set:
                    idx += 1
                    continue

                stride = 0
                for j in range(idx + 1, len(all_global_reads)):
                    if (
                        all_global_reads[j] - all_global_reads[j - 1] == 1
                        and all_global_reads[j] in all_global_write_set
                    ):
                        stride += 1
                    else:
                        break
                if stride >= 5:
                    # got one region
                    section = self.project.loader.find_section_containing(starting_offset)
                    if section is not None:
                        initial_data = self.project.loader.memory.load(starting_offset, stride)
                        end_data = func_call.result_state.solver.eval(
                            func_call.result_state.memory.load(starting_offset, stride), cast_to=bytes
                        )
                        if initial_data != end_data and self._is_string_reasonable(end_data):
                            region_candidates.append((starting_offset, stride, end_data))
                    idx += stride
                else:
                    idx += 1

            if region_candidates:
                type2_candidates.append((func.addr, desc, region_candidates))

        return type2_candidates

    def _analyze_type2(
        self, func_addr: int, desc: StringDeobFuncDescriptor, table_addrs: set[int]  # pylint:disable=unused-argument
    ) -> set:
        """
        Analyze Type 2 string deobfuscation functions, determine the following information:

        - Functions that load deobfuscated strings

        :param func_addr:
        :param desc:
        :return:
        """

        cfg = self.kb.cfgs.get_most_accurate()
        if cfg is None:
            raise AngrAnalysisError("StringObfuscationFinder needs a CFG for the analysis")

        # for each string table address, we find its string loader function
        # an obvious candidate function is 0x140001b20
        xrefs = self.kb.xrefs
        string_loader_candidates = set()
        for table_addr in table_addrs:
            xref_set = xrefs.get_xrefs_by_dst(table_addr)
            block_addrs = {xref.block_addr for xref in xref_set}
            for block_addr in block_addrs:
                if block_addr is None:
                    continue
                node = cfg.get_any_node(block_addr)
                if node is not None:
                    callees = list(self.kb.functions.callgraph.successors(node.function_address))
                    if callees:
                        #  string loader function should not call anything else
                        continue
                    string_loader_candidates.add(node.function_address)

        return string_loader_candidates

    def _find_type3(self) -> list[tuple[int, StringDeobFuncDescriptor]]:
        # Type 3 string deobfuscation functions
        # - Uses a buffer in the stack frame of its parent function
        # - Before the call, the values in the buffer or the struct are initialized during runtime
        # - The entire call can be simulated (it does not involve any other functions that angr does not support or do
        #   not have a SimProcedure for)

        cfg = self.kb.cfgs.get_most_accurate()
        if cfg is None:
            raise AngrAnalysisError("StringObfuscationFinder needs a CFG for the analysis")

        functions = self.kb.functions
        callgraph_digraph = networkx.DiGraph(functions.callgraph)

        sorted_funcs = GraphUtils.quasi_topological_sort_nodes(callgraph_digraph)
        tree_has_unsupported_funcs = {}
        function_candidates = []

        if self._function_addrs_to_analyze is not None:
            sorted_funcs = [a for a in sorted_funcs if a in self._function_addrs_to_analyze]

        for func_addr in sorted_funcs:
            if functions.get_by_addr(func_addr).is_simprocedure:
                # is this a stub SimProcedure?
                hooker = self.project.hooked_by(func_addr)
                if hooker is not None and hooker.is_stub:
                    tree_has_unsupported_funcs[func_addr] = True
            else:
                # which functions does it call?
                callees = list(callgraph_digraph.successors(func_addr))
                if any(tree_has_unsupported_funcs.get(callee, False) is True for callee in callees):
                    tree_has_unsupported_funcs[func_addr] = True
                else:
                    function_candidates.append(functions.get_by_addr(func_addr))

        type3_functions = []

        for func in function_candidates:
            if not 1 <= len(func.block_addrs_set) < 14:
                continue

            # if it has a prototype recovered, it must have four arguments
            if func.prototype is not None and len(func.prototype.args) not in {3, 4}:
                continue

            # the function must call some other functions
            # if callgraph_digraph.out_degree[func.addr] == 0:
            #     continue

            # take a look at its call sites
            func_node = cfg.get_any_node(func.addr)
            if func_node is None:
                continue
            call_sites = cfg.get_predecessors(func_node, jumpkind="Ijk_Call")
            if not call_sites:
                continue

            # examine the first 100 call sites and see if any of them sets up enough constants
            valid = False
            for i in range(min(100, len(call_sites))):
                call_site_block = self.project.factory.block(call_sites[i].addr)
                if self._is_block_setting_constants_to_stack(call_site_block):
                    valid = True
                    break
            if not valid:
                continue

            # take a look at the content
            with self._resilience():
                # catch all exceptions
                dec = self.project.analyses.Decompiler(func, cfg=cfg, fail_fast=self._fail_fast)  # type:ignore
            if dec.codegen is None or not dec.codegen.text:
                continue

            if not self._like_type3_deobfuscation_function(dec.codegen.text):
                continue

            # examine the first 100 call sites and see if any of them returns a valid string
            valid = False
            guessed_size = False
            for i in range(min(100, len(call_sites))):
                call_site_block = self.project.factory.block(call_sites[i].addr)
                if not self._is_block_setting_constants_to_stack(call_site_block):
                    continue

                # simulate an execution to see if it really works
                data, guessed_size = self._type3_prepare_and_execute(
                    func.addr, call_sites[i].addr, call_sites[i].function_address, cfg  # type:ignore
                )
                if data is None:
                    continue
                if len(data) > 3:
                    consecutive_printable_strs = self._consecutive_printable_substrings(data, min_length=4)
                    if consecutive_printable_strs:
                        valid = True
                        break

            if valid:
                desc = StringDeobFuncDescriptor()
                desc.string_output_arg_idx = 0
                desc.string_length_arg_idx = 1 if not guessed_size else None
                desc.string_null_terminating = False
                type3_functions.append((func.addr, desc))

        return type3_functions

    def _analyze_type3(
        self, func_addr: int, desc: StringDeobFuncDescriptor  # pylint:disable=unused-argument
    ) -> dict[int, bytes]:
        """
        Analyze Type 3 string deobfuscation functions, determine the following information:

        - The call sites
        - For each call site, the actual de-obfuscated content (in bytes)

        Decompiler will output the following code:

        *ptr = strdup("The deobfuscated string");
        *(ptr+8) = the string length;

        :param func_addr:
        :param desc:
        :return:
        """

        cfg = self.kb.cfgs.get_most_accurate()
        if cfg is None:
            raise AngrAnalysisError("StringObfuscationFinder needs a CFG for the analysis")

        cfg_node = cfg.get_any_node(func_addr)
        if cfg_node is None:
            raise AngrAnalysisError(f"Cannot find the CFG node for function {func_addr:#x}")
        call_sites = cfg.get_predecessors(cfg_node)
        callinsn2content = {}

        if self._function_addrs_to_analyze is not None:
            call_sites = [cs for cs in call_sites if cs.function_address in self._function_addrs_to_analyze]

        for idx, call_site in enumerate(call_sites):
            _l.debug("Analyzing type 3 candidate call site %#x (%d/%d)...", call_site.addr, idx + 1, len(call_sites))
            assert call_site.function_address is not None and isinstance(call_site.function_address, int)
            data, _ = self._type3_prepare_and_execute(func_addr, call_site.addr, call_site.function_address, cfg)
            if data:
                callinsn2content[call_site.instruction_addrs[-1]] = data
            # print(hex(call_site.addr), data)

        return callinsn2content

    #
    # Type 1 helpers
    #

    @staticmethod
    def _like_type1_deobfuscation_function(code: str) -> bool:
        return bool("^" in code or ">>" in code or "<<" in code)

    #
    # Type 2 helpers
    #

    @staticmethod
    def _like_type2_deobfuscation_function(code: str) -> bool:
        return bool(
            ("^" in code or ">>" in code or "<<" in code) and ("do" in code or "while" in code or "for" in code)
        )

    #
    # Type 3 helpers
    #

    @staticmethod
    def _like_type3_deobfuscation_function(code: str) -> bool:
        has_bitwise_ops = "^" in code or ">>" in code or "<<" in code or "~" in code
        has_loops = "do" in code or "while" in code or "for" in code
        has_many_bitwise_ops = code.count("^") + code.count(">>") + code.count("<<") + code.count("~") > 5
        return has_bitwise_ops and (has_loops or has_many_bitwise_ops)

    def _type3_prepare_and_execute(
        self, func_addr: int, call_site_addr: int, call_site_func_addr: int, cfg
    ) -> tuple[bytes | None, bool]:
        blocks_at_callsite = [call_site_addr]

        # backtrack from call site to include all previous consecutive blocks
        while True:
            pred_and_jumpkinds = cfg.get_predecessors_and_jumpkinds(
                cfg.get_any_node(call_site_addr), excluding_fakeret=False
            )
            if len(pred_and_jumpkinds) == 1:
                pred, jumpkind = pred_and_jumpkinds[0]
                if (
                    cfg.graph.out_degree[pred] == 1
                    and pred.addr + pred.size == call_site_addr
                    and jumpkind == "Ijk_Boring"
                ):
                    blocks_at_callsite.insert(0, pred.addr)
                    call_site_addr = pred.addr
                    continue
            break

        # take a look at the call-site block to see what registers are used
        reg_reads = set()
        for block_addr in blocks_at_callsite:
            reg_collector = IRSBRegisterCollector(self.project)
            reg_collector.process(state=None, block=self.project.factory.block(block_addr))
            reg_reads |= set(reg_collector.reg_reads)

        # run constant propagation to track constant registers
        prop = self.project.analyses.Propagator(
            func=self.kb.functions.get_by_addr(call_site_func_addr),
            only_consts=True,
            do_binops=True,
            vex_cross_insn_opt=True,
            load_callback=None,
            cache_results=True,
            key_prefix="cfg_intermediate",
        )

        # execute the block at the call site
        state = self.project.factory.blank_state(
            addr=call_site_addr,
            add_options={ZERO_FILL_UNCONSTRAINED_REGISTERS, ZERO_FILL_UNCONSTRAINED_MEMORY},
        )
        # setup sp and bp, just in case
        state.regs._sp = 0x7FFF0000
        bp_set = False
        assert prop.model.input_states is not None
        prop_state = prop.model.input_states.get(call_site_addr, None)
        if prop_state is not None:
            for reg_offset, reg_width in reg_reads:
                if reg_offset == state.arch.sp_offset:
                    continue
                if reg_width < 8:
                    # at least a byte
                    continue
                con = prop_state.load_register(reg_offset, reg_width // 8)
                if isinstance(con, claripy.ast.Base) and con.op == "BVV" and isinstance(con.concrete_value, int):
                    state.registers.store(reg_offset, claripy.BVV(con.concrete_value, reg_width))
                    if reg_offset == state.arch.bp_offset:
                        bp_set = True
        if not bp_set:
            state.regs._bp = 0x7FFF3000
        simgr = self.project.factory.simgr(state)

        # step until the call instruction
        for idx, block_addr in enumerate(blocks_at_callsite):
            if idx == len(blocks_at_callsite) - 1:
                inst = self.project.factory.block(block_addr).instructions
                simgr.step(num_inst=inst - 1)
            else:
                simgr.step()
            if not simgr.active:
                return None, False

        in_state = simgr.active[0]

        cc_cls = default_cc(self.project.arch.name, self.project.simos.name)
        assert (
            cc_cls is not None
        ), f"Failed to obtain the default calling convention for {self.project.arch.name}-{self.project.simos.name}."
        cc = cc_cls(self.project.arch)
        cc.STACKARG_SP_BUFF = 0  # disable shadow stack space because the binary code already sets it if needed
        cc.STACK_ALIGNMENT = 1  # disable stack address aligning because the binary code already sets it if needed
        prototype_0 = SimTypeFunction([], SimTypePointer(pts_to=SimTypeBottom(label="void"))).with_arch(
            self.project.arch
        )
        callable_0 = self.project.factory.callable(
            func_addr,
            concrete_only=True,
            base_state=in_state,
            cc=cc,
            prototype=prototype_0,
            add_options={ZERO_FILL_UNCONSTRAINED_MEMORY, ZERO_FILL_UNCONSTRAINED_REGISTERS},
            step_limit=STEP_LIMIT_ANALYSIS,
        )

        try:
            ret_value = callable_0()
        except (AngrCallableMultistateError, AngrCallableError):
            return None, False

        out_state = callable_0.result_state

        # figure out what was written
        assert out_state is not None and ret_value is not None
        ptr = out_state.memory.load(ret_value, size=self.project.arch.bytes, endness=self.project.arch.memory_endness)
        if out_state.memory.load(ptr, size=4).concrete_value == 0:
            # fall back to using the return value as the pointer
            ptr = ret_value
        if out_state.memory.load(ptr, size=4).concrete_value == 0:
            # can't find a valid pointer
            return None, False

        size = out_state.memory.load(ret_value + 8, size=4, endness=self.project.arch.memory_endness)
        guessed_size = False
        if size.symbolic or size.concrete_value == 0 or size.concrete_value >= 1024:
            size = 64
            guessed_size = True
        # TODO: Support lists with varied-length elements
        data = out_state.memory.load(ptr, size=size, endness="Iend_BE")
        if data.symbolic:
            return None, False

        return out_state.solver.eval(data, cast_to=bytes), guessed_size

    @staticmethod
    def _is_block_setting_constants_to_stack(block, threshold: int = 5) -> bool:
        insn_setting_const_bytes = 0
        xmm_has_const = False
        for insn in block.capstone.insns:
            if insn.mnemonic.startswith("mov") and len(insn.operands) == 2:
                if (
                    insn.operands[0].type == capstone.x86.X86_OP_MEM
                    and insn.operands[0].mem.base in {capstone.x86.X86_REG_RSP, capstone.x86.X86_REG_RBP}
                    and insn.operands[1].type == capstone.x86.X86_OP_IMM
                ):
                    # mov [rsp|rbp + offset], imm
                    insn_setting_const_bytes += 1  # FIXME: How to get the size of the mov in capstone?
                if (
                    insn.operands[0].type == capstone.x86.X86_OP_REG
                    and insn.operands[0].reg in ALL_X64_XMM_REGS
                    and insn.operands[1].type == capstone.x86.X86_OP_MEM
                    and insn.operands[1].mem.base == capstone.x86.X86_REG_RIP
                ):
                    xmm_has_const = True
                if (
                    xmm_has_const
                    and insn.operands[0].type == capstone.x86.X86_OP_MEM
                    and insn.operands[0].mem.base in {capstone.x86.X86_REG_RSP, capstone.x86.X86_REG_RBP}
                    and insn.operands[1].type == capstone.x86.X86_OP_REG
                    and insn.operands[1].reg in ALL_X64_XMM_REGS
                ):
                    # mov [rsp|rbp + offset], xmm0 - 31
                    insn_setting_const_bytes += 16

        return insn_setting_const_bytes >= threshold

    @staticmethod
    def _is_string_reasonable(s: bytes) -> bool:
        # test if the string is printable and is free of nonsense characters

        # TODO: Ask a local LLM
        s = s.replace(b"\x00", b"")
        return all(chr(ch) in string.printable for ch in s)

    @staticmethod
    def _consecutive_printable_substrings(s: bytes, min_length: int = 3) -> list[bytes]:
        """
        Find all consecutive printable substrings in a string.
        """
        substrings = []
        current_substring = b""
        for ch in s:
            if chr(ch) in string.printable:
                current_substring += bytes([ch])
            else:
                if current_substring:
                    if len(current_substring) >= min_length:
                        substrings.append(current_substring)
                    current_substring = b""
        if current_substring:
            substrings.append(current_substring)
        return substrings


AnalysesHub.register_default("StringObfuscationFinder", StringObfuscationFinder)
